import torch
from deluca.lung.controllers.core import Controller


# generic PID controller
class PID(Controller):
    def __init__(self, K=[3, 4, 0], RC=0.5, trainable=False, **kwargs):
        self.init_K = self.tensor(K)

        self.K = self.init_K if not trainable else torch.nn.Parameter(self.init_K)
        self.RC = RC

        self.reset()

    def reset(self):
        self.coef = self.tensor([0.0, 0.0, 0.0])

    @property
    def P(self):
        return self.coef[0]

    @property
    def I(self):
        return self.coef[1]

    @property
    def D(self):
        return self.coef[2]

    def compute_action(self, state, t):
        err = torch.tensor(self.waveform.at(t)) - state
        dt = self.dt(t)
        decay = dt / (dt + self.RC)

        self.coef = self.coef + self.tensor(
            [
                err - self.P,
                decay * (err - self.I),
                decay * (err - self.P - self.D),
            ]
        )

        u_in = torch.clamp(torch.dot(self.K, self.coef), min=0.0, max=100.0)

        return (u_in, self.u_out(t))



import os
from enum import Enum
import numpy as np
import dill as pickle
import torch

from deluca.lung.utils.core import BreathWaveform
from deluca.lung.core import VentObj


ControllerRegistry = []


class Phase(Enum):
    RAMP_UP = 1
    PIP = 2
    RAMP_DOWN = 3
    PEEP = 4


class Controller(VentObj):
    def __new__(cls, *args, **kwargs):
        obj = VentObj.__new__(cls)
        obj.__setattr__("time", float("inf"))
        obj.__setattr__("waveform", BreathWaveform())

        obj.kwargs(**kwargs)

        return obj

    @classmethod
    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        if cls.__name__ not in [ctrl.__name__ for ctrl in ControllerRegistry]:
            ControllerRegistry.append(cls)

    def reset(self):
        pass

    def compute_action(self, state, t):
        pass

    def forward(self, state, t):
        u_in, u_out = self.compute_action(state, t)
        self.time = t

        torched = isinstance(u_out, torch.Tensor)
        decay = (u_out if not torched else u_out.item()) == 1
        if decay:
            u_in = self.waveform.decay(t)
            if torched:
                u_in = self.tensor(u_in)
        return u_in, u_out

    def u_out(self, t):
        phase = self.phase(t)
        u_out = np.zeros_like(phase)
        u_out[np.equal(phase, Phase.RAMP_DOWN.value)] = 1
        u_out[np.equal(phase, Phase.PEEP.value)] = 1

        return self.tensor(u_out)

    def dt(self, t):
        dt = max(0, t - self.time)
        return dt

    def cycle_phase(self, t):
        return t % self.waveform.period

    def phase(self, t):
        return self.waveform.phase(t)

    def train(
        self,
        sim,
        duration=3,
        dt=0.03,
        epochs=100,
        use_noise = False,
        optimizer=torch.optim.Adam,
        optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
        loss_fn=torch.nn.L1Loss,
        loss_fn_params={},
        scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
        scheduler_params={},
        use_tqdm=True,
        print_loss=1,
        shuffle=False,
        device="cpu",
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

        # TODO: handle device-awareness
        for epoch in range(epochs):
            self.reset()
            sim.reset()

            loss = torch.tensor(0.0, device=device, requires_grad=True)

            self.zero_grad()

            for t in tt:
                sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.), std=1.)
                pressure = sim.pressure
                u_in, u_out = self(pressure, t)
                sim(u_in, u_out, t)

                if u_out == 0:
                    loss = loss + loss_fn(torch.tensor(self.waveform.at(t)), pressure)

            loss.backward(retain_graph=True)
            optimizer.step()
            scheduler.step(loss)
            per_step_loss = loss / len(tt)
            losses.append(per_step_loss)
            if epoch % print_loss == 0:
                print(
                    f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                )

        return losses



import numpy as np
from deluca.lung.controllers.core import Controller
from deluca.lung.utils.core import BreathWaveform


class ResidualExplorer(Controller):
    def __init__(self, base_controller, waveform=BreathWaveform(), **kwargs):
        self.base_controller = base_controller
        self.prev_u_out = 1

        self.delta_range = (-30, 30)
        self.duration_range = (0.1, 0.5)

        self.boundary_prob = 0.15
        self.boundary_delta_range = (0, 100)
        self.boundary_duration_range = (0.05, 0.3)

        self.needs_refresh = False

        self.update_waveform(waveform)
        self.refresh_exploration_perturbation()

    def update_waveform(self, waveform):
        self.waveform = waveform
        self.base_controller.waveform = waveform

    def refresh_exploration_perturbation(self):
        if np.random.random() < self.boundary_prob:
            self.resample_boundary_delta()
        else:
            self.resample_delta()

    def resample_delta(self):
        # random triangular bump
        height = np.random.uniform(*self.delta_range)
        duration = np.random.uniform(*self.duration_range)
    
        self.set_delta(height, duration)

    def set_delta(self, height, duration):
        t_min, t_max = 0, self.waveform._keypoints[2]

        t_begin = np.random.uniform(t_min, t_max - duration)
        t_mid = t_begin + 0.5 * duration
        t_end = t_begin + duration

        self.xp = np.array([t_min, t_begin, t_mid, t_end, t_max])
        self.fp = np.array([0.0, 0.0, height, 0.0, 0.0])

    def resample_boundary_delta(self):
        # random lean-to triangular bump at inspiratory boundary
        height = np.random.uniform(*self.boundary_delta_range)
        duration = np.random.uniform(*self.boundary_duration_range)
    
        self.set_boundary_delta(height, duration)

    def set_boundary_delta(self, height, duration):
        t_min, t_max = 0, self.waveform._keypoints[2]

        self.xp = np.array([t_min, duration, t_max])
        self.fp = np.array([height, 0, 0])

    def compute_action(self, state, t):
        base_u_in, base_u_out = self.base_controller(state, t)
        base_u_in = base_u_in.detach().numpy()
        base_u_out = base_u_out.detach().numpy()

        if base_u_out == 1:  # if base exhales, just follow base
            self.u_in = base_u_in
            self.u_out = 1
            if self.needs_refresh:
                self.refresh_exploration_perturbation()
                self.needs_refresh = False
        else:  # if base inhales, add residual to cycle phase
            self.needs_refresh = True
            self.u_in = base_u_in + np.interp(self.cycle_phase(t), self.xp, self.fp)
            self.u_out = 0

        return (np.clip(self.u_in, 0, 100), self.u_out)



from deluca.lung.environments.core import Environment
from deluca.lung.environments._balloon_lung import BalloonLung
from deluca.lung.environments._delay_lung import DelayLung
from deluca.lung.environments._physical_lung import PhysicalLung
from deluca.lung.environments._stitched_sim import StitchedSim

__all__ = [
    "Environment",
    "BalloonLung",
    "DelayLung",
    "PhysicalLung",
    "StitchedSim",
]



import sys
import torch
import click
import datetime
import time
import numpy as np
import re
import os
import logging

from venti.controllers.core import Controller
from venti.utils import Analyzer
from venti.utils import BreathWaveform
from venti.experimental import run_controller
from venti.experimental.core import run_calibration
from venti.experimental.core import save_data_and_plot


logging.basicConfig(stream=sys.stdout, level=logging.INFO)
log = logging.getLogger(__name__)

pattern = re.compile(".+R(\d+)_C(\d+)_PEEP(\d+).*")
RUNDIR = os.path.realpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "runs"))


def runner(run, r, c, peep, t, dt, abort, calibrate):
    log.info(locals())
    waveforms = [BreathWaveform((peep, pip)) for pip in [10, 15, 20, 25, 30, 35]]

    try:
        # If we specify R, C, and PEEP and they don't match the current lung
        # settings, skip
        match = pattern.match(run)
        if match is not None and not np.all(
            [int(setting) for setting in match.groups()] == [r, c, peep]
        ):
            log.info(
                f"Skipping {run}; wrong lung settings (found {match.groups()}, expecting ({r}, {c}, {peep}))"
            )
            sys.exit(1)

        controller_path = os.path.join(run, "controller.pkl")
        controller = Controller.load(controller_path)
        controller(torch.tensor(0.0), 0)

        timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        run_name = os.path.basename(run.replace("./", ""))

        if calibrate:
            run_calibration(r, c, peep, run, name="calib1")

        link = f"https://github.com/MinRegret/venti/tree/main/runs/completed/{run_name}"

        results = []
        plots = []
        for waveform in waveforms:
            name = f"result_{str(int(np.max(waveform.fp)))}"

            result = run_controller(
                controller_path,
                r,
                c,
                peep,
                waveform,
                t,
                dt,
                abort,
                name=name,
            )

            results.append(result)
            plots.append(f"### {name}\n![{name}]({name}.png)")

        if calibrate:
            run_calibration(r, c, peep, run, name="calib2")

        with open(os.path.join(RUNDIR, "report.md"), "r") as f:
            report = f.read().format(**{"title": run_name, "plots": "\n".join(plots)})

        with open(os.path.join(run, "README.md"), "w") as f:
            f.write(report)

        row = [f"[{run_name}]({link})"]
        total_loss = 0
        for result in results:
            analyzer = Analyzer(result)
            loss = analyzer.default_metric()
            row.append(f"{loss:.2f}")
            total_loss += loss
        row.append(f"{total_loss / len(results):.2f}")
        row.extend([str(x) for x in [r, c, peep, timestamp]])
        row = f"|{'|'.join(row)}|\n"

        with open(os.path.join(RUNDIR, "README.md"), "a") as f:
            f.write(row)

    except Exception as e:
        log.error(e)
        sys.exit(2)


@click.command()
@click.argument("run", type=click.Path(exists=True))
@click.option("-R", type=int, default=50, help="R value for phsyical lung")
@click.option("-C", type=int, default=10, help="C value for phsyical lung")
@click.option("--PEEP", type=int, default=5, help="PEEP")
@click.option("-T", type=int, default=600, help="Default timesteps")
@click.option("--dt", type=float, default=0.03, help="Time to wait")
@click.option("--abort", type=float, default=60, help="Abort pressure")
@click.option("--calibrate/--no-calibrate", default=True, help="Run calibration routine?")
@click.option("--sleep", type=int, default=60, help="Default time to sleep before trying again")
def _runner(run, r, c, peep, t, dt, abort, calibrate, sleep):
    runner(run, r, c, peep, t, dt, abort, calibrate)


if __name__ == "__main__":
    _runner()



import torch
import numpy as np

from deluca.lung.environments.core import Environment


def PropValve(x):  # copied from Controller.__SimulatedPropValve
    y = 3 * x
    flow_new = 1.0 * (torch.tanh(0.03 * (y - 130)) + 1)
    return torch.clamp(flow_new, 0.0, 1.72)


def Solenoid(x):  # copied from Controller.__SimulatedSolenoid
    if x > 0:
        return x / x
    else:
        return x * 0.


# balloon physics vent ported from Cohen lab's repository
# Sources:
# https://github.com/CohenLabPrinceton/deluca.lunglator-Dev/blob/master/sandbox/HOWTO_RunController.ipynb
# https://github.com/CohenLabPrinceton/deluca.lunglator-Dev/blob/master/vent/controller/control_module.py
class BalloonLung(Environment):
    def __init__(self, min_volume=6, peep_valve=5, PC=40, RP=1, leak=False, **kwargs):
        # dynamics hyperparameters
        self.min_volume = self.tensor(min_volume)
        self.PC = self.tensor(PC)
        self.RP = self.tensor(RP)
        self.P0 = self.tensor(0.0)
        self.leak = leak
        self.peep_valve = self.tensor(peep_valve)

        self.r0 = (3 * self.min_volume / (4 * np.pi)) ** (1 / 3)
        # reset states
        self.reset()

    def reset(self):
        # keep volume as the only free parameter
        self.volume = self.min_volume
        self.compute_aux_states()

        self.pips, self.peeps = [], []
        self.volumes, self.pressures = [], []

    def compute_aux_states(self):
        # compute all other state vars, which are just functions of volume
        r = (3 * self.volume / (4 * np.pi)) ** (1 / 3)
        self.pressure = self.P0 + self.PC * (1 - (self.r0 / r) ** 6) / (self.r0 ** 2 * r)

    def forward(self, u_in, u_out, t):
        self.pips.append(u_in)
        self.peeps.append(u_out)

        dt = self.dt

        # 2-dimensional action per timestep: PIP/PEEP voltages

        flow = torch.clamp(PropValve(u_in), 0, 2) * self.RP
        if self.pressure > self.peep_valve:
            flow = flow - torch.clamp(Solenoid(u_out), 0, 2) * 0.05 * self.pressure

        # update by flow rate
        self.volume = self.volume + flow * dt

        # simulate leakage
        if self.leak:
            RC = 5
            s = dt / (RC + dt)
            self.volume = self.volume + s * (self.min_volume - self.volume)

        # compute and record state
        self.compute_aux_states()
        self.volumes.append(self.volume)
        self.pressures.append(self.pressure)



import torch
import numpy as np
import sklearn.preprocessing
import matplotlib.pyplot as plt

from deluca.lung.utils.analyzer import Analyzer
from deluca.lung.core import VentObj

def resample(tt, f, dt=0.03, snap_left=False, return_grid=False):
    # resample a continuous-time function interpolated from non-uniform samples
    num_tt = int((tt[-1] - tt[0]) / dt)
    tt_grid = np.arange(num_tt)*dt + tt[0]

    if snap_left:
        f_grid = f[np.searchsorted(tt, tt_grid)]
    else:
        f_grid = np.interp(tt_grid, tt, f)

    if return_grid:
        return tt_grid, f_grid
    else:
        return f_grid

class Munger(VentObj):
    def __init__(self, paths, to_round=False, dt_resample=False, clip=(2, -1), **kwargs):
        self.paths = paths
        self.analyzers = []
        self.data = []
        self.splits = {}
        self.to_round = to_round

        # Add data
        self.add_data(paths, dt_resample, clip, **kwargs)

        # Shuffle and split data
        self.split_data(**kwargs)

        # Compute mean, std for u_in and pressure
        self.u_scaler, self.p_scaler = self.fit_scalers(**kwargs)

    # Note: the following methods are meant to be run in order
    def add_data(self, paths, dt_resample=True, clip=(2, -1), **kwargs):
        if isinstance(paths, str):
            paths = [paths]

        failed = 0
        for path in paths:
            try:
                analyzer = Analyzer(path)

                self.analyzers.append(analyzer)
                inspiratory_clips = analyzer.infer_inspiratory_phases()

                for start, end in inspiratory_clips[clip[0]:clip[1]]:  # skip first 2 & last breaths

                    if self.to_round:
                        u_in = np.round(analyzer.u_in[start:end])
                    else:
                        u_in = analyzer.u_in[start:end]

                    pressure = analyzer.pressure[start:end]

                    if dt_resample:
                        tt = analyzer.tt[start:end]
                        u_in = resample(tt, u_in)
                        pressure = resample(tt, pressure)

                    self.data.append((u_in, pressure))

            except Exception as e:
                print(path, e)
                failed += 1

        print(f"Added {len(self.data)} breaths from {len(self.paths) - failed} paths.")

    def split_data(self, seed=0, keys=["train", "test"], splits=[0.9, 0.1], **kwargs):
        rng = np.random.default_rng(seed)

        # Determine split boundaries
        splits = (np.array(splits) / np.sum(splits) * len(self.data)).astype("int")[:-1]

        # Everyday I'm shuffling
        rng.shuffle(self.data)

        # Splitting
        self.splits = {key: val for key, val in zip(keys, np.split(self.data, splits))}

    def fit_scalers(self, key="train"):
        u_scaler = sklearn.preprocessing.StandardScaler()
        u_scaler.fit(np.concatenate([u for u, p in self.splits[key]]).reshape(-1, 1))
        print(f"u_in: mean={u_scaler.mean_}, std={u_scaler.scale_}")

        p_scaler = sklearn.preprocessing.StandardScaler()
        p_scaler.fit(np.concatenate([p for u, p in self.splits[key]]).reshape(-1, 1))
        print(f"pressure: mean={p_scaler.mean_}, std={p_scaler.scale_}")

        return u_scaler, p_scaler

    def scale_and_window(self, key, u_window=5, p_window=3):
        X, y = [], []
        for u, p in self.splits[key]:
            T = len(u)
            u_scaled = self.u_scaler.transform(u.reshape(-1, 1))
            p_scaled = self.p_scaler.transform(p.reshape(-1, 1))

            for t in range(max(u_window, p_window) + 1, T):
                features = np.concatenate(
                    [u_scaled[t - u_window : t, 0], p_scaled[t - p_window : t, 0]]
                )
                target = p_scaled[t, 0]
                X.append(features)
                y.append(target)

        return torch.tensor(X), torch.tensor(y)

    def scale_and_window_boundary(self, key, boundary_index):
        X, y = [], []

        if boundary_index == 0:  # special case: no features, predict p[0]
            for u, p in self.splits[key]:
                p_scaled = self.p_scaler.transform([[p[0]]])
                target = p_scaled[0][0]
                y.append(target)
            return None, torch.tensor(y)

        # otherwise, collate [first B inputs, first B pressures] -> (next pressure) pairs
        for u, p in self.splits[key]:
            T = len(u)
            if T < boundary_index + 1:  # if trajectory is too short, abort
                continue

            u_scaled = self.u_scaler.transform(u[:boundary_index].reshape(-1, 1)).flat
            p_scaled = self.p_scaler.transform(p[: boundary_index + 1].reshape(-1, 1)).flat

            features = np.concatenate([u_scaled, p_scaled[:-1]])
            target = p_scaled[-1]

            X.append(features)
            y.append(target)

        return torch.tensor(X), torch.tensor(y)

    def get_data_loader(self, key, u_window=5, p_window=3, batch_size=512, shuffle=True):
        X, y = self.scale_and_window(key, u_window, p_window)
        dataset = torch.utils.data.TensorDataset(X, y)
        return torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)

    def get_boundary_data_loader(self, key, boundary_index, batch_size=512, shuffle=True):
        X, y = self.scale_and_window_boundary(key, boundary_index)
        dataset = torch.utils.data.TensorDataset(X, y)
        return torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle)

    def unscale_pressures(self, p):
        if type(p) is float:
            p = np.array([p])
        return self.p_scaler.inverse_transform(p.reshape(-1, 1))[:, 0]

    ###########################################################################
    # Plotting methods
    ###########################################################################

    def plot_boundary_pressures(self):
        plt.rc("figure", figsize=(16, 4))

        for tau in range(1, 6):
            plt.subplot(150 + tau)

            u_init = []
            p_init = []

            for u, p in self.splits["train"]:
                u_init.append(u[:tau].mean())
                p_init.append(p[tau])

            plt.xlim([0, 105])
            plt.ylim([2, 33])
            plt.xlabel(f"u_in[0:{tau}].mean")
            plt.title(f"pressure[{tau}]")

            plt.scatter(u_init, p_init, s=1)



import torch

from deluca.lung import Controller, Environment


class PIDCorrection(Controller):
    def __init__(self, base_controller: Controller, sim: Environment, pid_K=[0.0, 0.0], decay=0.1, **kwargs):
        self.base_controller = base_controller
        self.sim = sim
        self.I = 0.0
        self.K = pid_K
        self.decay = decay

        self.reset()

    def reset(self):
        self.base_controller.reset()
        self.sim.reset()
        self.I = 0.0

    def compute_action(self, state, t):
        u_in_base, u_out = self.base_controller(state, t)

        err = self.sim.pressure - state
        self.I = self.I * (1 - self.decay) + err * self.decay

        pid_correction = self.K[0] * err + self.K[1] * self.I

        u_in = torch.clamp(u_in_base + pid_correction, min=0.0, max=100.0)
        self.sim(u_in, u_out, t)

        return u_in, u_out



import os
import time
import datetime
import tqdm
import dill as pickle
import torch
import glob
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

import logging

from deluca.lung.core import VentObj
from deluca.lung.utils.analyzer import Analyzer
from deluca.lung.utils.munger import Munger

from deluca.lung.controllers._impulse import Impulse
from deluca.lung.controllers._predestined import Predestined
from deluca.lung.controllers._pid import PID

from deluca.lung.utils.core import BreathWaveform

EnvironmentRegistry = []

logger = logging.getLogger(__name__)


class Environment(VentObj):
    def __new__(cls, *args, **kwargs):
        obj = VentObj.__new__(cls)
        obj.__setattr__("pressure", 0)
        obj.__setattr__("flow", 0)
        obj.__setattr__("tt", 0)
        obj.__setattr__("dt", 0.03)

        obj.kwargs(**kwargs)

        return obj

    @classmethod
    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        if cls.__name__ not in [env.__name__ for env in EnvironmentRegistry]:
            EnvironmentRegistry.append(cls)

    # Define overrideable properties for run function. We could omit the
    # underscore, but since run is a function we won't touch and Environment
    # writers will likely want to use self.pressure as a float, we do this way
    @property
    def _pressure(self):
        return self.pressure

    @property
    def _flow(self):
        return self.flow

    @property
    def time(self):
        return self.tt * self.dt

    @property
    def is_real(self):
        return False

    def reset(self):
        self.pressure = 0
        self.flow = 0
        self.tt = 0

    def get_start_time(self):
        return 0

    def should_abort(self, pressure, flow, timestamp):
        return False

    def wait(self, duration):
        pass

    def forward(self, u_in, u_out, t):
        raise NotImplementedError()

    def run_cleanup(self):
        pass

    def run(
        self,
        controller,
        R=20,
        C=20,
        PEEP=5,
        T=3000,
        abort=None,
        sleep=None,
        directory=None,
        device="cpu",
        dt=None,
        use_tqdm=True,
    ):
        # Backwards compatibility...
        if abort is not None:
            self.abort = abort
        if dt is not None:
            self.dt = dt
        if PEEP is not None:
            self.PEEP = PEEP
        if sleep is not None:
            self.sleep = sleep

        result = {
            "controller": controller,
            "R": R,
            "C": C,
            "PEEP": self.PEEP,
            "dt": self.dt,
            "abort": self.abort,
            "T": T,
            "sleep": sleep,
            "directory": directory,
        }

        try:
            self.reset()
            controller.reset()

            self.tt = 0
            tt = range(T)

            timestamps = np.zeros(T)
            pressures = np.zeros(T)
            flows = np.zeros(T)
            u_ins = np.zeros(T)
            u_outs = np.zeros(T)

            begin = self.get_start_time()
            timestamp = begin

            pressure = 0

            if use_tqdm:
                tt = tqdm.tqdm(tt, leave=False)

            for i, t in enumerate(tt):
                pressure = self._pressure
                flow = self._flow

                prev_timestamp = timestamp
                timestamp = self.time - begin

                if self.should_abort(pressure, flow, timestamp):
                    break

                if device is not None and self.is_real:
                    pressure = torch.tensor(pressure, device=device)

                u_in, u_out = controller(pressure, timestamp)

                if device is not None and self.is_real:
                    u_in = u_in.item()
                    u_out = u_out.item()

                self(u_in, u_out, timestamp)

                timestamps[i] = timestamp
                pressures[i] = pressure
                flows[i] = flow
                u_ins[i] = u_in
                u_outs[i] = u_out

                duration = self.time - timestamp - begin
                self.wait(max(self.dt - duration, 0))

                self.tt += 1
        finally:
            self.run_cleanup()

        timeseries = {
            "timestamp": np.array(timestamps),
            "pressure": np.array(pressures),
            "flow": np.array(flows),
            "target": controller.waveform.at(timestamps),
            "u_in": np.array(u_ins),
            "u_out": np.array(u_outs),
        }

        for key, val in timeseries.items():
            timeseries[key] = val[: self.tt + 1]

        result["timeseries"] = timeseries

        if directory is not None:
            if not os.path.exists(directory):
                os.makedirs(directory)
            timestamp = datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
            pickle.dump(result, open(f"{directory}/{timestamp}.pkl", "wb"))

        return result

    def plot_impulse_responses(
        self,
        impulses=np.arange(0, 101, 10),
        zero=5,
        start=0.5,
        end=0.65,
        ylim=100,
        dt=0.03,
        T=100,
        use_tqdm=False,
        abort=70,
        **kwargs,
    ):
        analyzers = []

        for impulse in impulses:
            ir = Impulse(impulse, start, end)

            run_data = self.run(ir, dt=dt, T=T, PEEP=zero, use_tqdm=use_tqdm, abort=abort, **kwargs)
            analyzer = Analyzer(run_data)
            analyzers.append(analyzer)

        loss = 0
        colors = plt.cm.winter(np.linspace(0, 1, len(analyzers)))
        for i, analyzer in enumerate(analyzers):
            plt.plot(analyzer.tt, analyzer.pressure, color=colors[-i - 1])
            if impulses[0] == 0 and i == 0:
                loss += np.abs(analyzer.pressure - zero).mean()

        print(f"MAE for zero response: {loss}")

        plt.ylim(0, ylim)
        plt.show()

        return analyzers

    def plot_pids(self):
        plt.rc("figure", figsize=(8, 2))

        for coeff in np.linspace(0.0, 0.5, 20):
            pid = PID([coeff, 0.5 - coeff, 0], waveform=BreathWaveform((5, 35)))
            self.reset()
            result = self.run(pid, T=333, abort=100)
            analyzer = Analyzer(result)

            cmap = matplotlib.cm.get_cmap("rainbow")
            plt.plot(analyzer.tt, analyzer.pressure, "b", c=cmap(2 * coeff))



from deluca.lung.controllers.core import Controller
from deluca.lung.utils import BreathWaveform

import numpy as np
import torch

class ClippedAdvDeep(Controller):
    def __init__(
        self,
        clip=35.0,
        H=100,
        waveform=None,
        bptt=1,
        input_dim=1,
        activation=torch.nn.ReLU,
        history_len=15,
        kernel_size=5,
        **kwargs
    ):
        super().__init__()

        self.model = torch.nn.Sequential(
            torch.nn.Conv1d(input_dim, H, kernel_size),
            activation(),
            torch.nn.Flatten(),
            torch.nn.Linear(H * (history_len - kernel_size + 1), 1),
        )

        # linear feature transform:
        # errs -> [average of last h errs, ..., average of last 2 errs, last err]
        # emulates low-pass filter bank

        self.featurizer = torch.ones((2*history_len, history_len), requires_grad=False).tril_()
        self.featurizer /= torch.arange(history_len, 0, -1).unsqueeze(0)

        self.history_len = history_len
        self.input_dim = input_dim
        self.waveform = waveform or BreathWaveform()

        self.clip = clip

        self.reset()

    def reset(self):
        self.errs = [self.tensor(0.0)] * self.history_len
        self.pressures = [self.tensor(0.0)] * self.history_len
        self.targets = [self.tensor(0.0)] * self.history_len

    def update(self, key, state):
        getattr(self, key).append(state)

    def compute_action(self, state, t):
        target = self.tensor(self.waveform.at(t))
        self.errs.append(target - state)
        self.pressures.append(state)
        self.targets.append(target)

        decay = self.waveform.decay(t)

        if decay is None:
            trajectory = torch.stack(self.errs[-self.history_len :] + self.pressures[-self.history_len :]).unsqueeze(0).unsqueeze(0)
            u_in = self.model(trajectory @ self.featurizer)
        else:
            u_in = self.tensor(decay)
            self.errs = [self.tensor(0.0)] * self.history_len
            self.pressures = [self.tensor(0.0)] * self.history_len
            self.targets = [self.tensor(0.0)] * self.history_len

        u_in = torch.clamp(u_in, min=0.0, max=self.clip).squeeze()

        return (u_in, self.u_out(t))



import torch
import numpy as np

from deluca.lung.environments.core import Environment


# 2nd attempt at vent simulation
# observed pressure comes from delayed control signal, with back-pressure from deluca.lung
class DelayLung(Environment):
    def __init__(self, delay=10, **kwargs):
        # dynamics hyperparameters
        self.params = {
            "min_volume": 1.5,
            "R_lung": 10,
            "C_lung": 6,
            "inertia": 0.995,
            "control_gain": 0.02,
        }
        self.delay = delay
        self.params = {key: self.tensor(val) for key, val in self.params.items()}
        if kwargs is not None:
            for p in kwargs:
                self.params[p] = kwargs[p]

        # reset states
        self.r0 = (3 * self.params["min_volume"] / (4 * np.pi)) ** (1 / 3)
        self.reset()

    def reset(self):
        self.volume = self.tensor(self.params["min_volume"])
        self.pipe_pressure = self.tensor(0.0)

        self.controls_in, self.controls_out = [], []
        self.volumes, self.pressures = [], []
        self.compute_aux_states()

    def compute_aux_states(self):
        # compute all other state vars, which are just functions of volume
        r = (3 * self.volume / (4 * np.pi)) ** (1 / 3)
        self.vent_pressure = self.params["C_lung"] * (1 - (self.r0 / r) ** 6) / (self.r0 ** 2 * r)

        if len(self.controls_in) < self.delay:
            self.pipe_impulse = 0
            self.peep = 0
        else:
            self.pipe_impulse = self.params["control_gain"] * self.controls_in[-self.delay]
            self.peep = self.controls_out[-self.delay]

        self.pipe_pressure = self.params["inertia"] * self.pipe_pressure + self.pipe_impulse
        self.pressure = torch.max(self.tensor(0.0), self.pipe_pressure - self.vent_pressure)

        if self.peep:
            self.pipe_pressure *= 0.995

    def forward(self, u_in, u_out, t):
        u_in = max(0, u_in)

        self.controls_in.append(u_in)
        self.controls_out.append(u_out)

        dt = self.dt

        # 2-dimensional action per timestep
        flow = self.pressure / self.params["R_lung"]

        # update by flow rate
        self.volume = torch.max(self.volume + flow * dt, self.params["min_volume"])

        # compute and record state
        self.compute_aux_states()
        self.pressures.append(self.pressure)
        self.volumes.append(self.volume)



import numpy as np
from deluca.lung.controllers.core import Controller


class Explorer(Controller):
    def __init__(self, period=0.5, u_in_max=5, p_switch=0.1, max_inhale=2.5):
        self.period = period
        self.u_in_max = u_in_max
        self.p_switch = p_switch
        self.max_inhale = max_inhale

        self.prev_cycle = -1
        self.last_exhale_time = 0

        self.u_out = 0

    def compute_action(self, state, t):
        cur_cycle = int(t / self.period)
        if self.prev_cycle != cur_cycle:
            self.prev_cycle = cur_cycle
            self.u_in = np.random.randint(0, self.u_in_max + 1)

            if np.random.random() < self.p_switch:
                self.u_out = 1 - self.u_out

            if state > 45:
                self.u_out = 1

        if self.u_out == 1:
            self.last_exhale_time = t
        else:
            if t - self.last_exhale_time > self.max_inhale:
                self.u_out = 1
                self.last_exhale_time = t

        return (self.u_in, self.u_out)



import time
import numpy as np
import datetime

from deluca.lung.environments.core import Environment


class PhysicalLung(Environment):
    def __init__(
        self,
        hal=None,
        sleep=3.0,
        abort=50,
        PEEP=5,
        dt_threshold=1.25,
        dt_patience=10,
        peep_threshold=0.5,
        peep_patience=10,
        peep_breaths=2,
    ):
        if hal is None:
            from deluca.lung import Hal

            hal = Hal()
        self.hal = hal
        self.sleep = sleep

        self.abort = abort
        self.PEEP = PEEP
        self.dt_threshold = dt_threshold
        self.dt_patience = dt_patience
        self.peep_threshold = peep_threshold
        self.peep_patience = peep_patience

        self.breaths = 0
        self.peep_breaths = peep_breaths

        self.reset()

    def reset(self):
        self.prev_timestamp = float("-inf")
        self.dt_window = np.ones(self.dt_patience) * self.dt
        self.pressure_window = np.ones(self.peep_patience) * self.PEEP
        self.breaths = 0

    @property
    def _pressure(self):
        return self.hal.pressure

    @property
    def time(self):
        return time.time()

    @property
    def is_real(self):
        return True

    def get_start_time(self):
        return datetime.datetime.now().timestamp()

    def should_abort(self, pressure, flow, timestamp):
        if pressure > self.abort:
            print(f"Pressure of {pressure} > {self.abort}; quitting")
            return True

        self.dt_window = np.roll(self.dt_window, 1)
        self.dt_window[0] = timestamp - max(self.dt, self.prev_timestamp)
        self.prev_timestamp = timestamp

        if np.mean(self.dt_window) > self.dt * self.dt_threshold:
            # print(
                # f"dt averaged {100 * self.dt_threshold:.1f}% higher over the last {self.dt_patience} timesteps; quitting"
            # )
            return False

        if self.breaths > self.peep_breaths:
            self.pressure_window = np.roll(self.pressure_window, 1)
            self.pressure_window[0] = pressure

        if np.mean(self.pressure_window) < self.PEEP * self.peep_threshold:
            print("Pressure drop, did you blow up?")
            return True

        return False

    def wait(self, duration):
        time.sleep(duration)

    def forward(self, u_in, u_out, t):
        if u_out == 1:
            self.breaths += 1
        self.hal.setpoint_in = u_in
        self.hal.setpoint_ex = u_out

    def run_cleanup(self):
        self.hal.setpoint_in = 0
        self.hal.setpoint_ex = 1
        time.sleep(self.sleep)
        self.hal.setpoint_ex = 0



import deluca.lung
from deluca.lung.utils import Analyzer

import os
import tqdm
import torch
import numpy as np
import itertools

DEFAULT_DATA_DIR = f"{deluca.lung.ROOT}/data"


def where(arr, idxs):
    return [arr[idx] for idx in idxs]


def best_pid(
    Rs=[50],
    Cs=[10],
    PEEPs=[5],
    directory=DEFAULT_DATA_DIR,
    loss_fn=np.abs,
    k=10,
    p_mask=lambda x: True,
    i_mask=lambda x: True,
    d_mask=lambda x: x == 0,
):
    losses = {}
    for R in Rs:
        for C in Cs:
            for PEEP in PEEPs:
                pid_dir = os.path.join(directory, f"R{R}_C{C}_PEEP{PEEP}", "pid")
                for path in tqdm.tqdm(os.listdir(pid_dir)):
                    if path[-4:] == ".pkl":
                        try:
                            path = os.path.join(pid_dir, path)
                            analyzer = Analyzer(path)

                            K = analyzer.controller.K
                            if isinstance(K, torch.Tensor):
                                K = K.detach().numpy()
                            K = tuple(K.tolist())
                            if not (p_mask(K[0]) and i_mask(K[1]) and d_mask(K[2])):
                                continue
                            if K not in losses:
                                losses[K] = []
                            loss = analyzer.default_metric(loss_fn=loss_fn)
                            losses[K].append(loss)
                        except Exception as e:
                            print(path, e)

    keys = list(losses.keys())
    losses = [np.mean(losses[key]) for key in keys]
    k = min(k, len(keys))
    idxs = np.argsort(losses)

    return where(keys, idxs[:k]), where(losses, idxs[:k])


def global_best_pid(directory=DEFAULT_DATA_DIR, loss_fn=np.abs, k=10):
    pkls = []
    losses = {}
    for (R, C) in itertools.product([5,20,50], [10,20,50]):
        pid_dir = os.path.join(directory, f"R{R}_C{C}_PEEP{5}", "pid")
        for path in tqdm.tqdm(os.listdir(pid_dir)):
            if path[-4:] == ".pkl":

                try:

                    path = os.path.join(pid_dir, path)
                    analyzer = Analyzer(path)

                    pkls.append(path)
                    K = analyzer.controller.K
                    if isinstance(K, torch.Tensor):
                        K = K.detach().numpy()
                    K = tuple(K.tolist())
                    if K not in losses:
                        losses[K] = []
                    loss = analyzer.default_metric(loss_fn=loss_fn)
                    losses[K].append(loss)
                except Exception as e:
                    print(path, e)
    keys = list(losses.keys())
    losses = [np.mean(losses[key]) for key in keys]
    k = min(k, len(keys))
    idxs = np.argsort(losses)

    return where(keys, idxs[:k]), where(losses, idxs[:k])


def plot_pid(R=50, C=10, PEEP=5, K=(2.0, 10.0, 0.0), directory=DEFAULT_DATA_DIR):
    pid_dir = os.path.join(directory, f"R{R}_C{C}_PEEP{PEEP}", "pid")
    K = np.array(K)

    for path in os.listdir(pid_dir):
        if path[-4:] == ".pkl":
            try:
                path = os.path.join(pid_dir, path)
                analyzer = Analyzer(path)

                CK = analyzer.controller.K
                if isinstance(CK, torch.Tensor):
                    CK = CK.detach().numpy()

                if np.all(np.array(K) == CK):
                    analyzer.plot(
                        title=f"Range: ({analyzer.controller.waveform.PEEP}, {analyzer.controller.waveform.PIP}), {path}"
                    )
            except Exception as e:
                print(path, e)



import numpy as np
from deluca.lung.controllers.core import Controller


class PeriodicImpulse(Controller):
    def __init__(self, period=2, duration=0.1, hold=0.5, amplitude=20, hold_amplitude=0):
        self.period = period
        self.duration = duration
        self.hold = hold
        self.amplitude = amplitude
        self.hold_amplitude = hold_amplitude

    def compute_action(self, state, t):
        phase = t % self.period

        if phase < self.duration:
            u_in = self.amplitude
            u_out = 0
        elif phase < self.hold:
            u_in = self.hold_amplitude
            u_out = 0
        else:
            u_in = 0
            u_out = 1

        return (u_in, u_out)



import numpy as np
from deluca.lung.controllers.core import Controller


class SpikyExplorer(Controller):
    def __init__(
        self,
        period=0.5,
        u_in_max=5,
        p_switch=0.1,
        max_inhale=2.5,
        spike_duration=0.1,
        spike_prob=0.2,
        spike_min=20,
    ):
        self.period = period
        self.u_in_max = u_in_max
        self.p_switch = p_switch
        self.max_inhale = max_inhale
        self.spike_duration = spike_duration
        self.spike_prob = spike_prob
        self.spike_min = spike_min

        self.prev_cycle = -1
        self.last_exhale_time = 0
        self.last_cycle_start = 0
        self.spike_value = 0

        self.u_out = 0

    def compute_action(self, state, t):
        cur_cycle = int(t / self.period)

        if self.prev_cycle != cur_cycle:
            self.prev_cycle = cur_cycle
            self.last_cycle_start = t

            self.u_in = np.random.randint(0, self.u_in_max + 1)

            if np.random.random() < self.p_switch:
                self.u_out = 1 - self.u_out

            if np.random.random() < self.spike_prob and state < self.spike_min:
                self.spike_value = np.random.randint(self.u_in_max + 1, 101)
            else:
                self.spike_value = self.u_in

        if self.u_out == 1:
            self.last_exhale_time = t
        else:
            if t - self.last_exhale_time > self.max_inhale:
                self.u_out = 1
                self.last_exhale_time = t

        if t - self.last_cycle_start < self.spike_duration:
            u_in = self.spike_value
        else:
            u_in = self.u_in

        u_out = self.u_out

        if state > 45:  # safety trigger until next phase
            u_in = self.u_in = self.spike_value = 0
            u_out = self.u_out = 1

        return (u_in, u_out)



import numpy as np
import tqdm

from deluca.lungollers import PID
from deluca.lungonments import PhysicalLung
from deluca.lung import Analyzer, BreathWaveform

from deluca.lungimental.core import run_calibration

DEFAULT_VALUES = np.concatenate((np.linspace(0, 0.9, 10), np.linspace(1, 10, 10)))


def pid_grid(
    Ps=DEFAULT_VALUES,
    Is=DEFAULT_VALUES,
    Ds=[0.0],
    PIPs=[10, 15, 20, 25, 30, 35],
    R=50,
    C=10,
    PEEP=5,
    abort=60,
    T=300,
    directory=None,
    vent=None,
    **kwargs,
):
    analyzers = []
    vent = vent or PhysicalLung()

    print("Running calibration")
    run_calibration(R, C, PEEP, directory)

    print("Running grid")
    for PIP in tqdm.tqdm(PIPs):
        for P in tqdm.tqdm(Ps, leave=False):
            for I in tqdm.tqdm(Is, leave=False):
                for D in tqdm.tqdm(Ds, leave=False):
                    vent.reset()

                    waveform = BreathWaveform((PEEP, PIP))
                    pid = PID(K=[P, I, D], waveform=waveform)
                    analyzer = vent.run(
                        pid,
                        R=R,
                        C=C,
                        PEEP=PEEP,
                        abort=abort,
                        directory=directory,
                        T=T,
                        use_tqdm=False,
                        **kwargs,
                    )
                    analyzers.append(analyzer)

    print("Running calibration")
    run_calibration(R, C, PEEP, directory)

    return analyzers



from deluca.lung.controllers._pid import PID
from deluca.lung.environments._physical_lung import PhysicalLung
from deluca.lung.utils.analyzer import Analyzer

import pickle
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt


def save_data_and_plot(result, directory, name):
    pickle.dump(result, open(f"{directory}/{name}.pkl", "wb"))
    Analyzer(result).plot()
    plt.savefig(f"{directory}/{name}.png")


def run_calibration(R, C, PEEP, directory, name=None, T=300):
    vent = PhysicalLung()
    pid = PID(K=[1.0, 0.0, 0.0])
    result = vent.run(
        pid, R=R, C=C, PEEP=PEEP, T=T, directory=f"{directory}/calibration", abort=60
    )

    if name is not None:
        save_data_and_plot(result, directory, name)

    return result



from abc import ABC, abstractmethod
from deluca.lung.devices.pins import Pin, PWMOutput

import numpy as np


class SolenoidBase(ABC):
    """ An abstract baseclass that defines methods using valve terminology.
    Also allows configuring both normally _open and normally closed valves (called the "form" of the valve).
    """
    _FORMS = {'Normally Closed': 0,
              'Normally Open': 1}

    def __init__(self, form='Normally Closed'):
        """

        Args:
            form (str): The form of the solenoid; can be either `Normally Open` or `Normally Closed`
        """
        self.form = form

    @property
    def form(self) -> str:
        """ Returns the human-readable form of the valve."""
        return dict(map(reversed, self._FORMS.items()))[self._form]

    @form.setter
    def form(self, form):
        """ Performs validation on requested form and then sets it.

        Args:
            form (str): The form of the solenoid; can be either `Normally Open` or `Normally Closed`
        """
        if form not in self._FORMS.keys():
            raise ValueError('form must be one of {}'.format(self._FORMS.keys()))
        else:
            self._form = self._FORMS[form]

    @abstractmethod
    def open(self):
        """ Energizes valve if Normally Closed. De-energizes if Normally Open."""

    @abstractmethod
    def close(self):
        """ De-energizes valve if Normally Closed. Energizes if Normally Open."""

    @property
    @abstractmethod
    def is_open(self) -> bool:
        """ Returns True if valve is open, False if it is closed"""


class OnOffValve(SolenoidBase, Pin):
    """ An extension of vent.io.iobase.Pin which uses valve terminology for its methods.
    Also allows configuring both normally _open and normally closed valves (called the "form" of the valve).
    """
    _FORMS = {'Normally Closed': 0,
              'Normally Open': 1}

    def __init__(self, pin, form='Normally Closed', gpio=None):
        """

        Args:
            pin (int): The number of the pin to use
            form (str): The form of the solenoid; can be either `Normally Open` or `Normally Closed`
            gpio (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        self.form = form
        Pin.__init__(self, pin, gpio)
        SolenoidBase.__init__(self, form=form)

    def open(self):
        """ Energizes valve if Normally Closed. De-energizes if Normally Open."""
        if self._form:
            self.write(0)
        else:
            self.write(1)

    def close(self):
        """ De-energizes valve if Normally Closed. Energizes if Normally Open."""
        if self.form == 'Normally Closed':
            self.write(0)
        else:
            self.write(1)

    @property
    def is_open(self) -> bool:
        """ Implements parent's abstractmethod; returns True if valve is open, False if it is closed"""
        energized = True if self.read() else False
        if self.form == 'Normally Closed':
            return energized
        else:
            return not energized


class PWMControlValve(SolenoidBase, PWMOutput):
    """ An extension of PWMOutput which incorporates linear
    compensation of the valve's response.
    """

    def __init__(self, pin, form='Normally Closed', frequency=None, response=None, gpio=None):
        """
        Args:
            pin (int): The number of the pin to use
            form (str): The form of the solenoid; can be either `Normally Open` or `Normally Closed`
            frequency (float): The PWM frequency to use.
            response (str): "/path/to/response/curve/file"
            gpio (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        PWMOutput.__init__(self, pin=pin, initial_duty=0, frequency=frequency, gpio=gpio)
        SolenoidBase.__init__(self, form=form)
        '''if response is None:
            raise NotImplementedError('You need to implement a default response behavior')'''
        if form != 'Normally Closed':
            raise NotImplementedError('Normally Open PWM control valves have not been implemented')
        self._rising = True
        self._load_valve_response(response_path=response)

    @property
    def is_open(self) -> bool:
        """ Implements parent's abstractmethod; returns True if valve is open, False if it is closed"""
        if self.setpoint > 0:
            return True
        else:
            return False

    def open(self):
        """ Implements parent's abstractmethod; fully opens the valve"""
        self.setpoint = 1.0

    def close(self):
        """ Implements parent's abstractmethod; fully closes the valve"""
        self.setpoint = 0.0

    @property
    def setpoint(self) -> float:
        """ The linearized setpoint corresponding to the current duty cycle according to the valve's response curve

        Returns:
            float: A number between 0 and 1 representing the current flow as a proportion of maximum
        """
        return self.inverse_response(self.duty, self._rising)

    @setpoint.setter
    def setpoint(self, setpoint):
        """Overridden to determine & write the duty cycle corresponding
        to the requested linearized setpoint according to the valve's
        response curve

        Args:
            setpoint (float): A number between 0 and 100 representing how much to open the valve
        """
        if not 0 <= setpoint <= 100:
            raise ValueError('setpoint must be between 0 and 100 for an expiratory control valve')
        self._rising = setpoint > self.setpoint
        self.duty = self.response(setpoint, self._rising)

    def response(self, setpoint, rising=True):
        """Setpoint takes a value in the range (0,100) so as not to confuse with duty cycle, which takes a value in the
        range (0,1). Response curves are specific to individual valves and are to be implemented by subclasses.
        Different curves are calibrated to 'rising = True' (valves opening) or'rising = False' (valves closing), as
        different characteristic flow behavior can be observed.

        Args:
            setpoint (float): A number between 0 and 1 representing how much to open the valve
            rising (bool): Whether or not the requested setpoint is higher than the last (rising = True), or the
                opposite (Rising = False)

        Returns:
            float: The PWM duty cycle corresponding to the requested setpoint
        """

        idx = (np.abs(self._response_array[:, 0] - setpoint)).argmin()
        if rising:
            duty = self._response_array[idx, 1]
        else:
            duty = self._response_array[idx, 2]

        return duty

    def inverse_response(self, duty_cycle, rising=True):

        """Inverse of response. Given a duty cycle in the range (0,1), returns the corresponding linear setpoint in the
        range (0,100).

        Args:
            duty_cycle: The PWM duty cycle
            rising (bool): Whether or not the requested setpoint is higher than the last (rising = True), or the
                opposite (Rising = False)

        Returns:
            float: The setpoint of the valve corresponding to `duty_cycle`
        """
        if rising:
            idx = (np.abs(self._response_array[:, 1] - duty_cycle)).argmin()
        else:
            idx = (np.abs(self._response_array[:, 2] - duty_cycle)).argmin()
        return self._response_array[idx, 0]

    def _load_valve_response(self, response_path):
        """ Loads and applies a response curve of the form `f(setpoint) = duty`. A response curve maps the underlying
        PWM duty cycle `duty` onto the normalized variable `setpoint` representing the flow through the valve as a
        percentage of its maximum.

        Flow through a proportional valve may be nonlinear with respect to [PWM] duty cycle, if the valve itself does
        not include its own electronics to linearize response wrt/ input. Absent on-board compensation of response, a
        proportional solenoid with likely not respond [flow] at all below some minimum threshold duty cycle.
        Above this threshold, the proportional valve begins to open and its response resembles a sigmoid: just past the
        threshold there is a region where flow increases exponentially wrt/ duty cycle, this is followed by a region of
        pseudo-linear response that begins to taper off, eventually approaching the valve's maximum flow asymptotically
        as the duty cycle approaches 100% and the valve opens fully.

        Args:
            response_path: 'path/to/binary/response/file' - if response_path is None, defaults to `setpoint = duty`
        """
        if response_path is not None:
            response_array = np.load(response_path)
        else:
            response_array = np.linspace([0, 0, 0], [100, 1, 1], num=101)
        self._response_array = response_array



from deluca.lung.controllers.core import Controller
from deluca.lung.environments._physical_lung import PhysicalLung
from deluca.lung.utils.core import BreathWaveform
from deluca.lung.experimental.core import save_data_and_plot

import os
import click
import pickle


def run_controller(path, R, C, PEEP, waveform, T, dt, abort, name=None):
    directory = os.path.dirname(os.path.realpath(path))
    experiment = os.path.basename(directory)
    pickle.dump(locals(), open(os.path.join(directory, "meta.pkl"), "wb"))

    controller = Controller.load(path)

    if waveform is not None:
        controller.waveform = waveform
        controller.reset()

    lung = PhysicalLung()

    result = lung.run(controller, R=R, C=C, PEEP=PEEP, T=T, dt=dt, abort=abort)

    if name is not None:
        save_data_and_plot(result, directory, name)

    return result


@click.command()
@click.argument("path", type=click.Path(exists=True), required=1)
@click.option("-R", type=int, default=50, help="R value for phsyical lung")
@click.option("-C", type=int, default=10, help="C value for phsyical lung")
@click.option("--PEEP", type=int, default=5, help="PEEP")
@click.option("--waveform", type=BreathWaveform, default=None, help="Waveform")
@click.option("-T", type=int, default=1000, help="Default timesteps")
@click.option("--dt", type=float, default=0.03, help="Time to wait")
@click.option("--abort", type=float, default=60, help="Abort pressure")
def _run_controller(path, r, c, peep, waveform, t, dt, abort):
    return run_controller(path, r, c, peep, waveform, t, dt, abort)


if __name__ == "__main__":
    _run_controller()



import os
import torch
import dill as pickle


class VentObj(torch.nn.Module):
    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj.__setattr__("name", cls.__name__)
        torch.nn.Module.__init__(obj)
        return obj

    def kwargs(self, **kwargs):
        for kw, arg in kwargs.items():
            self.__setattr__(kw, arg)

    def save(self, path, thing=None):
        obj = thing or self
        dirname = os.path.abspath(os.path.dirname(path))
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        pickle.dump(obj, open(path, "wb"))

    def tensor(self, val, requires_grad=True, dtype=None, device="cpu"):
        if dtype is None:
            dtype = torch.get_default_dtype()
        return torch.tensor(val, requires_grad=requires_grad, dtype=dtype, device=device)

    @classmethod
    def load(cls, path):
        return pickle.load(open(path, "rb"))



from deluca.lung.controllers.core import Controller
from deluca.lung.controllers._pid import PID
from deluca.lung.controllers._explorer import Explorer
from deluca.lung.controllers._impulse import Impulse
from deluca.lung.controllers._predestined import Predestined
from deluca.lung.controllers._periodic_impulse import PeriodicImpulse
from deluca.lung.controllers._spiky_explorer import SpikyExplorer
from deluca.lung.controllers._residual_explorer import ResidualExplorer


__all__ = [
    "Controller",
    "PID",
    "Explorer",
    "Impulse",
    "Predestined",
    "PeriodicImpulse",
    "SpikyExplorer",
    "ResidualExplorer"
]



from deluca.lung.devices import I2CDevice, be16_to_native
from abc import ABC, abstractmethod
import random
from collections import deque

import time
import numpy as np


class Sensor(ABC):
    """ Abstract base Class describing generalized sensors. Defines a mechanism for limited internal storage of recent
    observations and methods to pull that data out for external use.
    """
    _DEFAULT_STORED_OBSERVATIONS = 128

    def __init__(self):
        """ Upon creation, calls update() to ensure that if get is called there will be something to return."""
        self._maxlen_data = self._DEFAULT_STORED_OBSERVATIONS
        self._data = {
            'timestamp': deque(maxlen=self.maxlen_data),
            'value': deque(maxlen=self.maxlen_data)
        }

    def update(self) -> float:
        """ Make a sensor reading, verify that it makes sense and store the result internally. Returns True if reading
        was verified and False if something went wrong.
        """
        value = self._read()
        self._data['timestamp'].append(time.time())
        self._data['value'].append(value)
        return self._verify(value)

    def get(self, average=False) -> float:
        """ Return the most recent sensor reading, or an average of readings since last get(). Clears internal memory so as not to have stale data."""
        # FIXME - timeout decorators for everyone
        if len(self._data['value']) == 0:
            self.update()
        if average:
            value = np.mean(self._data['value'])
        else:
            value = self._data['value'].pop()
        self._clear()
        return value

    def age(self) -> float:
        """ Returns the time in seconds since the last sensor update, or -1 if never updated."""
        if not self._data['timestamp']:
            return -1.0
        else:
            return time.time() - self._data['timestamp'][-1]

    def reset(self):
        """ Resets the sensors internal memory. May be overloaded by subclasses to extend functionality specific to a
        device.
        """
        self._clear()
            
    def _clear(self):
        """ Resets the sensors internal memory. """
        for key in self._data.keys():
            self._data[key].clear()

    @property
    def data(self) -> np.array:
        """ Returns all Locally-stored observations. 

        Returns:
            np.array: An array of timestamped observations arranged oldest to newest.
        """
        result = np.column_stack([self._data['timestamp'], self._data['value']])
        self._clear()
        return result

    @property
    def maxlen_data(self) -> int:
        """ Returns the number of observations kept in the Sensor's internal ndarray. Once the ndarray has been filled,
        the sensor begins overwriting the oldest elements of the ndarray with new observations such that the size of the
        internal storage stays constant.
        """
        return self._maxlen_data

    @maxlen_data.setter
    def maxlen_data(self, new_data_length):
        """ Set a new length for stored observations. Clears existing
        observations and resets.

        Args:
            new_data_length (int): The new length of internal observation storage
        """
        if type(new_data_length) != int:
            raise ValueError
        new_data = {'value': deque(maxlen=new_data_length), 'timestamp': deque(maxlen=new_data_length)}
        for key in new_data.keys():
            new_data[key].extend(self._data[key])
            self._data[key].clear()
            self._data[key] = new_data[key]
        self._maxlen_data = new_data_length

    def _read(self) -> float:
        """ Calls _raw_read and scales the result before returning it."""
        return self._convert(self._raw_read())

    @abstractmethod
    def _verify(self, value):
        """ Validate reading and throw exception/alarm if sensor does not appear to be working correctly.
        """

    @abstractmethod
    def _convert(self, raw):
        """ Converts a raw reading from a sensor in whatever format the device communicates with into a meaningful
        result.
        """

    @abstractmethod
    def _raw_read(self):
        """ Requests a new observation from the device and returns the raw result in whatever format/units the device
        communicates with.
        """


class AnalogSensor(Sensor):
    """ Generalized class describing an analog sensor attached to the ADS1115 analog to digital converter. Inherits from
    the sensor base class and extends with functionality specific to analog sensors attached to the ADS1115. If
    instantiated without a subclass, conceptually represents a voltmeter with a normalized output.
    """
    # TODO The offset voltage/output span & verify/calibrate stuff needs a rethink.
    _DEFAULT_offset_voltage = 0
    _DEFAULT_output_span = 5
    _DEFAULT_CALIBRATION = {
        'offset_voltage': _DEFAULT_offset_voltage,
        'output_span': _DEFAULT_output_span,
        'conversion_factor': 1
    }

    def __init__(self, adc, **kwargs):
        """ Links analog sensor on the ADC with configuration options specified. If no options are specified, it assumes
        the settings currently on the ADC.

        Args:
            adc (vent.io.devices.ADS1115): The adc object to which the AnalogSensor is attached
            **kwargs: `field=value` - see vent.io.devices.ADS1115 for additional documentation. Strongly suggested to
                specify `MUX=adc_pin` here unless you know what you're doing.
        """
        super().__init__()
        self.adc = adc
        if 'MUX' not in (kwargs.keys()):
            raise TypeError(
                'User must specify MUX for AnalogSensor creation'
            )
        kwargs = {key: kwargs[key] for key in kwargs.keys() - ('gpio',)}
        self._check_and_set_attr(**kwargs)

    def calibrate(self, **kwargs):
        """ Sets the calibration of the sensor, either to the values contained in the passed tuple or by some routine;
        the current routine is pretty rudimentary and only calibrates offset voltage.

        Args:
            **kwargs: calibration_field=value, where calibration field is one of the following: 'offset_voltage',
                output_span' or 'conversion_factor'
        """
        # FIXME
        if kwargs:
            for fld, val in kwargs.items():
                if fld in self._DEFAULT_CALIBRATION.keys():
                    setattr(self, fld, val)
        else:
            for _ in range(50):
                self.update()
                # PRINT FOR DEBUG / HARDWARE TESTING
                print(
                    "Analog Sensor Calibration @ {:6.4f}".format(self.data[self.data.shape[0] - 1]),
                    end='\r'
                )
                time.sleep(.1)
            self.offset_voltage = np.min(self.data[-50:])
            # PRINT FOR DEBUG / HARDWARE TESTING
            print("Calibrated low-end of AnalogSensor @",
                  ' %6.4f V' % self.offset_voltage)

    def _read(self) -> float:
        """ Returns a value in the range of 0 - 1 corresponding to a fraction of the full input range of the sensor."""
        return self._convert(self._raw_read())

    def _verify(self, value) -> bool:
        """ Checks to make sure sensor reading was indeed in [0, 1].

        Args:
            value (float): Sensor reading to validate
        """
        report = bool(-1 <= value / self.conversion_factor <= 1)
       # if not report:
            # FIXME: Right now this just expands the calibration range whenever bounds are exceeded, because we're not
            #  familiar enough with our sensors to know when we should really be rejecting values. This approach should
            #  work for debugging/R&D purposes but really ought to be thought through and/or replaced for production.
            #  For example, negative voltages are probably bad. voltages about VDD (~5V) are also probably bad. There is
            #  some expected drift around offset voltage and output span, however, and that drift is going to change
            #  depending on the sensor in question; i.e., voltages between offset_voltage and zero may or may not be ok,
            #  and voltages above the offset+span that do not exceed VDD may or may not be ok as well.
        #    self.offset_voltage = min(self.offset_voltage, value)
        #    self.output_span = max(self.output_span, value - self.offset_voltage)
        #    print('Warning: AnalogSensor calibration adjusted')
        return report

    def _convert(self, raw) -> float:
        """ Scales raw voltage into the range 0 - 1.

        Args:
            raw (float): The raw sensor reading to convert.
        """
        return (
                self.conversion_factor * ((raw - getattr(self, 'offset_voltage')) / getattr(self, 'output_span'))
        )

    def _raw_read(self) -> float:
        """ Builds kwargs from configured fields to pass along to adc, then calls adc.read_conversion(), which returns
        a raw voltage.
        """
        fields = self.adc.USER_CONFIGURABLE_FIELDS
        kwargs = dict(zip(
            fields,
            (getattr(self, field) for field in fields)
        ))
        return self.adc.read_conversion(**kwargs)

    def _fill_attr(self):
        """ Examines self to see if there are any fields identified as user configurable or calibration that have not
        been write (i.e. were not passed to __init__ as **kwargs). If a field is missing, grabs the default value either
        from the ADC or from _DEFAULT_CALIBRATION and sets it as an attribute.
        """
        for cfld in self.adc.USER_CONFIGURABLE_FIELDS:
            if not hasattr(self, cfld):
                setattr(
                    self,
                    cfld,
                    getattr(self.adc.config, cfld).unpack(self.adc.cfg)
                )
        for dcal, value in self._DEFAULT_CALIBRATION.items():
            if not hasattr(self, dcal):
                setattr(self, dcal, value)

    def _check_and_set_attr(self, **kwargs):
        """ Checks to see if arguments passed to __init__ are recognized as user configurable or calibration fields. If
        so, write the value as an attribute like: self.KEY = VALUE. Keeps track of how many attributes are write in this
        way; if at the end there unknown arguments leftover, raises a TypeError; otherwise, calls _fill_attr() to fill
        in fields that were not passed as arguments.

        Args:
            **kwargs: `field=value` - see vent.io.devices.ADS1115 for additional documentation
        """
        allowed = (
            *self.adc.USER_CONFIGURABLE_FIELDS,
            *self._DEFAULT_CALIBRATION.keys(),
        )
        result = 0
        for fld, val in kwargs.items():
            if fld in allowed:
                setattr(self, fld, val)
                result += 1
        if result != len(kwargs):
            raise TypeError('AnalogSensor was passed unknown field(s)', kwargs.items(), allowed)
        self._fill_attr()


class SFM3200(Sensor, I2CDevice):
    """ I2C Inspiratory flow sensor manufactured by Sensirion AG. Range: +/- 250 SLM
    Datasheet:
         https://www.sensirion.com/fileadmin/user_upload/customers/sensirion/Dokumente/ ...
            ... 5_Mass_Flow_Meters/Datasheets/Sensirion_Mass_Flow_Meters_SFM3200_Datasheet.pdf
    """
    _DEFAULT_ADDRESS = 0x40
    _FLOW_OFFSET = 32768
    _FLOW_SCALE_FACTOR = 120

    def __init__(self, address=_DEFAULT_ADDRESS, i2c_bus=1, gpio=None):
        """
        Args:
            address (int): The I2C Address of the SFM3200 (usually 0x40)
            i2c_bus (int): The I2C Bus to use (usually `1` on the Raspberry Pi)
            gpio (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        I2CDevice.__init__(self, address, i2c_bus, gpio)
        Sensor.__init__(self)
        self.reset()
        self._start()

    def reset(self):
        """ Extended to add device specific behavior: Asks the sensor to perform a soft reset. 80 ms soft reset time."""
        super().reset()
        self.write_device(0x2000)
        time.sleep(.08)  # TODO: this should be an await

    def _start(self):
        """ Device specific:Sends the 'start measurement' command to the sensor. Start-up time once command has been
        recieved is 'less than 100ms'
        """
        self.write_device(0x1000)
        time.sleep(.1) # TODO: this should be an await

    def _verify(self, value) -> bool:
        """ No further verification needed for this sensor. Onboard chip handles all that. Could throw in a CRC8 checker
        instead of discarding them in _convert().

        Args:
            value (float): The sensor reading to verify.
        """
        return True

    def _convert(self, raw) -> float:
        """ Overloaded to replace with device-specific protocol. Convert raw int to a flow reading having type float
        with units slm. Implementation differs from parent for clarity and consistency with source material.

        Source:
          https://www.sensirion.com/fileadmin/user_upload/customers/sensirion/Dokumente/ ...
            ... 5_Mass_Flow_Meters/Application_Notes/Sensirion_Mass_Flo_Meters_SFM3xxx_I2C_Functional_Description.pdf

        Args:
            raw (int): The raw value read from the SFM3200
        """
        return (raw - self._FLOW_OFFSET) / self._FLOW_SCALE_FACTOR

    def _raw_read(self) -> int:
        """ Performs an read on the sensor, converts received bytearray, discards the last two bytes (crc values - could
        implement in future), and returns a signed int converted from the big endian two complement that remains.
        """
        return be16_to_native(self.read_device(4))



import torch

from deluca.lung.controllers import PID
from deluca.lung.controllers.core import Controller
from deluca.lung.utils import BreathWaveform
import itertools
import numpy as np
import random
from scipy.special import softmax


class DeepPIDResidual(Controller):
    def __init__(
            self,
            H=100,
            waveform=None,
            bptt=1,
            input_dim=1,
            activation=torch.nn.ReLU,
            history_len=10,
            kernel_size=5,
            normalize=False,
            time_as_feature=False,
            u_scaler=None,
            p_scaler=None,
            pid_K=[3.0, 4.0, 0.0],
            **kwargs
    ):
        super().__init__()

        self.model = torch.nn.Sequential(
            torch.nn.Conv1d(input_dim, H, kernel_size),
            activation(),
            torch.nn.Flatten(),
            torch.nn.Linear(H * (history_len - kernel_size + 1), 1),
        )

        self.time_as_feature = time_as_feature
        if self.time_as_feature:
            self.time_model = torch.nn.Sequential(torch.nn.Linear(1, 10), activation(), torch.nn.Linear(10, 1))

        multiplier = self.tensor(1.0)
        self.residual_multiplier = torch.nn.Parameter(multiplier)

        # linear feature transform:
        # errs -> [average of last h errs, ..., average of last 2 errs, last err]
        # emulates low-pass filter bank

        self.featurizer = torch.ones((history_len, history_len), requires_grad=False).tril_()
        self.featurizer /= torch.arange(history_len, 0, -1).unsqueeze(0)

        self.history_len = history_len
        self.input_dim = input_dim
        self.waveform = waveform or BreathWaveform()
        self.pid_base = PID(K=pid_K)
        self.normalize = normalize
        if normalize:
            self.u_scaler = u_scaler
            self.p_scaler = p_scaler

        self.reset()

    def reset(self):
        self.errs = [self.tensor(0.0)] * self.history_len
        self.pid_base.waveform = self.waveform
        self.pid_base.reset()

    def update(self, key, state):
        getattr(self, key).append(state)

    def compute_action(self, state, t):
        self.pid_base.waveform = self.waveform
        target = self.tensor(self.waveform.at(t))
        if self.normalize:
            target_scaled = self.p_scaler.transform(target).squeeze()
            state_scaled = self.p_scaler.transform(state).squeeze()
            self.errs.append(target_scaled - state_scaled)
        else:
            self.errs.append(target - state)

        decay = self.waveform.decay(t)

        if decay is None:
            trajectory = torch.stack(self.errs[-self.history_len:]).unsqueeze(0).unsqueeze(0)
            u_in_base, _ = self.pid_base.compute_action(state, t)
            u_in_residual = self.model(trajectory @ self.featurizer)
            if self.time_as_feature:
                u_in_residual += self.time_model(self.tensor(self.cycle_phase(t)).unsqueeze(0))
            # u_in = u_in_base + u_in_residual * 0.1 * torch.relu(torch.tanh(5e-2 * self.residual_multiplier))
            u_in = u_in_base + u_in_residual * 1e-3
        else:
            u_in = self.tensor(decay)

        u_in = torch.clamp(u_in, min=0.0, max=100.0).squeeze()

        return (u_in, self.u_out(t))

    def train(
            self,
            sim,
            pip_feed="parallel",
            duration=3,
            dt=0.03,
            epochs=100,
            use_noise=False,
            optimizer=torch.optim.Adam,
            optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
            loss_fn=torch.nn.L1Loss,
            loss_fn_params={},
            scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
            scheduler_params={"factor": 0.9, "patience": 10},
            use_tqdm=True,
            print_loss=1,
            shuffle=False,
            device="cpu",
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

        PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5

        # TODO: handle device-awareness
        for epoch in range(epochs):

            if pip_feed == "parallel":
                self.zero_grad()
                loss = torch.tensor(0.0, device=device, requires_grad=True)

            for PIP in random.sample(PIPs, 6):

                if pip_feed == "sequential":
                    self.zero_grad()
                    loss = torch.tensor(0.0, device=device, requires_grad=True)

                self.waveform = BreathWaveform((PEEP, PIP))

                self.reset()
                sim.reset()

                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.5), std=1.)
                    pressure = sim.pressure
                    u_in, u_out = self(pressure, t)
                    sim(u_in, u_out,
                        t)  # potentially add multiplicative noise by * torch.normal(mean=torch.tensor(1.5), std=0.5)

                    if u_out == 0:
                        loss = loss + loss_fn(torch.tensor(self.waveform.at(t)), pressure)

                if pip_feed == "sequential":
                    loss.backward(retain_graph=True)
                    optimizer.step()
                    scheduler.step(loss)
                    per_step_loss = loss / len(tt)
                    losses.append(per_step_loss)

                    if epoch % print_loss == 0:
                        print(
                            f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                        )

            if pip_feed == "parallel":
                loss.backward(retain_graph=True)
                optimizer.step()
                scheduler.step(loss)
                per_step_loss = loss / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                    )

        return losses

    def train_global(
            self,
            sims,
            pip_feed="parallel",
            duration=3,
            dt=0.03,
            epochs=100,
            use_noise=False,
            optimizer=torch.optim.Adam,
            optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
            loss_fn=torch.nn.L1Loss,
            loss_fn_params={},
            scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
            scheduler_params={"factor": 0.9, "patience": 10},
            use_tqdm=True,
            print_loss=1,
            shuffle=False,
            device="cpu", 
        PIPs = [10, 15, 20, 25, 30, 35],
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

#         PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5

        # TODO: handle device-awareness
        for epoch in range(epochs):

            if pip_feed == "parallel":
                self.zero_grad()
                loss = torch.tensor(0.0, device=device, requires_grad=True)

            for PIP, sim in itertools.product(PIPs, sims):

                if pip_feed == "sequential":
                    self.zero_grad()
                    loss = torch.tensor(0.0, device=device, requires_grad=True)

                self.waveform = BreathWaveform((PEEP, PIP))

                self.reset()
                sim.reset()

                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.5), std=1.)
                    pressure = sim.pressure
                    u_in, u_out = self(pressure, t)
                    sim(u_in, u_out,
                        t)  # potentially add multiplicative noise by * torch.normal(mean=torch.tensor(1.5), std=0.5)

                    if u_out == 0:
                        loss = loss + loss_fn(torch.tensor(self.waveform.at(t)), pressure)

                if pip_feed == "sequential":
                    loss.backward(retain_graph=True)
                    optimizer.step()
                    scheduler.step(loss)
                    per_step_loss = loss / len(tt)
                    losses.append(per_step_loss)

                    if epoch % print_loss == 0:
                        print(
                            f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                        )

            if pip_feed == "parallel":
                loss.backward(retain_graph=True)
                optimizer.step()
                scheduler.step(loss)
                per_step_loss = loss / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                    )

        return losses

    def train_global_boosted(
            self,
            sim_datas, #Make this a list of (R, C, sim)s
            duration=3,
            dt=0.03,
            epochs=100,
            use_noise=False,
            optimizer=torch.optim.Adam,
            optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
            loss_fn=torch.nn.L1Loss,
            loss_fn_params={},
            scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
            scheduler_params={"factor": 0.9, "patience": 1000},
            alpha=0.5,
            device="cpu",
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

        PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5
        PIPSims = list(itertools.product(PIPs, sim_datas))
        loss_by_simpip = np.zeros(len(PIPSims))
        self.zero_grad()

        idxs =range(len(PIPSims))
        for epoch in range(epochs):
            p = softmax(alpha * loss_by_simpip)
            loss = torch.tensor(0.0, device=device, requires_grad=True)
            for idx in idxs:
                PIP, (R, C, sim) = PIPSims[idx]
                self.waveform = BreathWaveform((PEEP, PIP))

                self.reset()
                sim.reset()
                loss_idx = 0.0
                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(1.5), std=1.)
                    pressure = sim.pressure
                    u_in, u_out = self(pressure, t)
                    sim(u_in, u_out, t)

                    if u_out == 0:
                        curr_loss = loss_fn(torch.tensor(self.waveform.at(t)), pressure)
                        loss = loss + curr_loss * p[idx]
                        loss_idx += curr_loss

                loss.backward(retain_graph=True)
                per_step_loss = loss_idx / len(tt)
                losses.append(per_step_loss)
                loss_by_simpip[idx] = per_step_loss
            optimizer.step()
            self.zero_grad()

            print(f"Epoch: {epoch}, PIP: {PIP} R: {R} C: {C}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}")
            for i in range(len(PIPSims)):
                PIP,(R,C, _) = PIPSims[i]
                print(f"PIP, {PIP}, R {R}, C {C}, loss {loss_by_simpip[i] :.2f}")
            print(f"loss mean {np.mean(loss_by_simpip):.2f}+-{np.std(loss_by_simpip):.2f}")
        return losses



from deluca.lung.utils.munger import Munger

import math
import torch
from collections import OrderedDict
import numpy as np
import matplotlib.pyplot as plt


class SNN(torch.nn.Module):
    def __init__(self, in_dim, out_dim, hidden_dim, n_layers, dropout_prob=0.0):
        super().__init__()
        layers = OrderedDict()
        for i in range(n_layers - 1):
            if i == 0:
                layers[f"fc{i}"] = torch.nn.Linear(in_dim, hidden_dim, bias=False)
            else:
                layers[f"fc{i}"] = torch.nn.Linear(hidden_dim, hidden_dim, bias=False)
            layers[f"selu_{i}"] = torch.nn.SELU()
            layers[f"dropout_{i}"] = torch.nn.AlphaDropout(p=dropout_prob)
        layers[f"fc_{i+1}"] = torch.nn.Linear(hidden_dim, out_dim, bias=True)
        self.network = torch.nn.Sequential(layers)
        self.reset_parameters()

    def forward(self, x):
        return self.network(x)

    def reset_parameters(self):
        for layer in self.network:
            if not isinstance(layer, torch.nn.Linear):
                continue
            torch.nn.init.normal_(layer.weight, std=1 / math.sqrt(layer.out_features))
            if layer.bias is not None:
                fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(layer.weight)
                bound = 1 / math.sqrt(fan_in)
                torch.nn.init.uniform_(layer.bias, -bound, bound)

    def track_layer_activations(self, x):
        activations = []
        for layer in self.network:
            x = layer.forward(x)
            if isinstance(layer, torch.nn.SELU):
                activations.append(x.data.flatten())
        return activations

    def train(
        self,
        munger_path,
        u_window=5,
        p_window=3,
        train_key="train",
        test_key="test",
        batch_size=512,
        epochs=500,
        optimizer=torch.optim.Adam,
        optimizer_params={"lr": 1e-3},
        scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
        scheduler_params={"factor":0.9, "patience":15},
        loss_fn=torch.nn.L1Loss,
        loss_fn_params={},
        print_loss=100,
    ):
        munger = Munger.load(munger_path)
        loader = munger.get_data_loader(
            key=train_key, u_window=u_window, p_window=p_window, batch_size=batch_size
        )
        X_tensor, y_tensor = munger.scale_and_window(
            key=train_key, u_window=u_window, p_window=p_window
        )
        X_test_tensor, y_test_tensor = munger.scale_and_window(
            key=test_key, u_window=u_window, p_window=p_window
        )

        optim = optimizer(self.parameters(), **optimizer_params)
        schedule = scheduler(optim, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        for epoch in range(epochs):
            for X_batch, y_batch in loader:
                self.zero_grad()
                preds = self(X_batch).squeeze()
                loss = loss_fn(preds, y_batch)
                loss.backward()
                optim.step()
            schedule.step(loss)

            if epoch % print_loss == 0:
                with torch.no_grad():
                    # expensive end-of-epoch eval, just for intuition
                    preds = self(X_tensor).squeeze()
                    train_loss = loss_fn(preds, y_tensor)

                    # cross-validation
                    preds = self(X_test_tensor).squeeze()
                    test_loss = loss_fn(preds, y_test_tensor)

                print(
                    f"Epoch {epoch:2d}: train={train_loss.item():.5f}, test={test_loss.item():.5f}"
                )

    def plot_residuals(self, munger_path, u_window=5, p_window=3, figsize=(10, 10), s=5, alpha=0.5):
        munger = Munger.load(munger_path)
        plt.rc('figure', figsize=figsize)

        X_test, y_test = munger.scale_and_window('test', u_window=u_window, p_window=p_window)

        with torch.no_grad():
            y_pred = self(X_test).numpy()

        unscaled_truth = munger.unscale_pressures(y_test)
        unscaled_preds = munger.unscale_pressures(y_pred)

        x = np.linspace(0, 45, 1000)

        plt.xlabel('true pressure')
        plt.ylabel('predicted pressure')

        plt.scatter(unscaled_truth, unscaled_preds, s=s, alpha=alpha)
        plt.plot(x, x, linestyle='-', color = "red", label = "f(x) = x");


class BoundaryDict(torch.nn.Module):
    def __init__(self, models=None, const_trainable=True):
        self.dict = {}

        if models is None:
            models = []

        models.insert(0, ConstantModel(trainable=const_trainable))

        for i, model in enumerate(models):
            self.dict[i] = model

        self.width = i

    def __getitem__(self, key):
        return self.dict[key]

    def __setitem__(self, key, val):
        self.dict[key] = val

    def __delitem__(self, key):
        del self.dict[key]

    def items(self):
        return self.dict.items()

    def train(
        self,
        munger_path,
        train_key="train",
        test_key="test",
        batch_size=16,
        epochs=800,
        optimizer=torch.optim.SGD,
        optimizer_params={"lr": 1e-2, "weight_decay": 1e-3},
        scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
        scheduler_params={"factor":0.9, "patience":15},
        loss_fn=torch.nn.L1Loss,
        loss_fn_params={},
        print_loss=100,
    ):
        munger = Munger.load(munger_path)
        loss_fn = loss_fn(**loss_fn_params)

        for idx, model in self.dict.items():
            if idx == 0:
                _, y = munger.scale_and_window_boundary(train_key, boundary_index=0)
                self[0].update_constant(y.mean().item())
                continue

            loader = munger.get_boundary_data_loader(train_key, idx, batch_size=batch_size)

            # TODO: figure out if we want to pull out 'test'
            X_test, y_test = munger.scale_and_window_boundary(test_key, idx)
            optim = optimizer(model.parameters(), **optimizer_params)
            schedule = scheduler(optim, **scheduler_params)

            if print_loss is not None:
                print("Boundary model #", idx)

            for epoch in range(epochs):
                for X_batch, y_batch in loader:
                    model.zero_grad()
                    preds = model(X_batch).squeeze()
                    loss = loss_fn(preds, y_batch)
                    loss.backward()
                    optim.step()
                schedule.step(loss)

                if epoch % print_loss == 0:
                    with torch.no_grad():
                        test_loss = loss_fn(model(X_test).squeeze(), y_test)

                    print(f"Epoch {epoch:2d}: test={test_loss.item():.5f}")

            self[idx] = model

    def plot_boundary_performance(self, munger_path, key="test"):
        munger = Munger.load(munger_path)

        plt.rc("figure", figsize=(16, 4))

        for boundary_index in range(1, self.width + 1):
            plt.subplot(2, 5, boundary_index)

            model = self[boundary_index]

            X_test, y_test = munger.scale_and_window_boundary(key, boundary_index)

            with torch.no_grad():
                y_pred = model(X_test).numpy()

            unscaled_truth = munger.unscale_pressures(y_test)
            unscaled_preds = munger.unscale_pressures(y_pred)

            x = np.linspace(0, 25, 1000)

            plt.xlabel("true pressure")
            plt.ylabel("predicted pressure")

            plt.scatter(
                unscaled_truth, unscaled_preds, s=5, label=f"Boundary Model #{boundary_index}"
            )
            plt.plot(x, x, linestyle="-", color="red", label="f(x) = x")


class ShallowBoundaryModel(torch.nn.Module):  # Paula's boundary model
    def __init__(self, in_dim, hidden_dim=100, out_dim=1):
        super().__init__()
        layers = OrderedDict()
        layers["fc1"] = torch.nn.Linear(in_dim, hidden_dim, bias=False)
        layers["tanh_1"] = torch.nn.Tanh()
        layers["fc_2"] = torch.nn.Linear(hidden_dim, out_dim, bias=True)
        self.model = torch.nn.Sequential(layers)

    def forward(self, x):
        out = self.model(x)
        return out


class RegressionBoundaryModel(torch.nn.Module):
    def __init__(self, in_dim, out_dim=1):
        super().__init__()
        layers = OrderedDict()
        layers["fc_1"] = torch.nn.Linear(in_dim, out_dim, bias=True)
        self.model = torch.nn.Sequential(layers)

    def forward(self, x):
        out = self.model(x)
        return out


class ConstantModel(torch.nn.Module):  # boundary model to predict first pressure with no features
    def __init__(self, const_out=0., trainable=True):
        super().__init__()

        # Assume we want to fix trainable at initialization
        self.trainable = trainable
        self.update_constant(const_out)

    def update_constant(self, const_out):
        # Note: we have this round-about logic b/c torch complains about
        # updating a parameter with a tensor value
        const_out = torch.tensor(const_out)
        if self.trainable:
            self.const_out = torch.nn.parameter.Parameter(const_out)
        else:
            self.const_out = const_out

    def forward(self, x=None):
        return self.const_out


class InspiratoryModel(
    torch.nn.Module
):  # manager of a family of trained boundary models + general model
    def __init__(self, default_model, boundary_dict):
        super().__init__()

        self.default_model = default_model
        self.boundary_dict = torch.nn.ModuleDict({str(i): m for i, m in boundary_dict.items()})

    def forward(self, t, features):  # placeholder for stitching
        if t in self.boundary_dict:
            return self.boundary_dict[str(t)](features)
        return self.default_model(features)



""" Base classes & functions used throughout vent.io.devices
"""
from collections import OrderedDict
import functools
import traceback

import time
import pigpio


def gpio_command(func):
    @functools.wraps(func)
    def exception_catcher(self, *args, **kwargs):
        result = None
        try:
            result = func(self, *args, **kwargs)
        except Exception as e:
            print(e, traceback.TracebackException.from_exception(e))
        return result

    return exception_catcher


class PigpioConnection(pigpio.pi):
    """ Subclass that extends pigpio.pi to throw an exception if there are issues connecting to the pigpio daemon."""

    def __init__(self, *args, **kwargs):
        """ Calls superclass init and checks if a connection was established; throws a RuntimeError if not.

        Args:
            *args: parameters to pass through like: pigpio.pi().__init__(*args)
            **kwargs: parameters to pass through like: pigpio.pi().__init__(**kwargs)
        """
        super().__init__(*args, **kwargs)
        if not self.connected:
            raise RuntimeError('Could not establish connection with pigpio daemon')


class IODeviceBase:
    """ Abstract base Class for pigpio handles (or whatever other GPIO library
    we end up using)

    Note: pigpio commands return -144 if an error is encountered while
    attempting to communicate with the demon. TODO would be to recognize
    when that occurs and handle it gracefully, i.e. kill the daemon,
    restart it, and reopen the python interface(s)
    """

    def __init__(self, gpio: PigpioConnection = None):
        """ Initializes the pigpio python bindings object if necessary,
        and checks that it is actually running.

        Args:
            gpio (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        self._gpio = gpio if gpio is not None else PigpioConnection(show_errors=False)
        self._handle = -1

    @property
    def gpio(self) -> PigpioConnection:
        """ The pigpio python bindings object"""
        return self._gpio

    @property
    def handle(self) -> int:
        """ Pigpiod handle associated with device (only for i2c/spi)"""
        return self._handle

    @property
    def pigpiod_ok(self) -> bool:
        """ Returns True if pigpiod is running and False if not"""
        return self.gpio.connected

    def _close(self):
        """ Closes an I2C/SPI (or potentially Serial) connection"""
        if not self.gpiopiod_ok or self.handle <= 0:
            return


class I2CDevice(IODeviceBase):
    """ A class wrapper for pigpio I2C handles. Defines several methods
    used for reading from and writing to device registers. Defines
    helper classes Register and ValueField for handling the
    manipulation of arbitrary registers.

    Note: The Raspberry Pi uses LE byte-ordering, while the outside
    world tends to use BE (at least, the sensors in use so far all do).
    Thus, bytes need to be swapped from native (LE) ordering to BE
    prior to being written to an i2c device, and bytes recieved need to
    be swapped from BE into native (LE). All methods except read_device
    and write_device perform this automatically. The methods read_device
    and write_device do NOT byteswap and return bytearrays rather than
    the unsigned 16-bit int used by the other read/write methods.
    """

    def __init__(self, i2c_address, i2c_bus, pig=None):
        """ Initializes pigpio bindings and opens i2c connection.

        Args:
            i2c_address (int): I2C address of the device. (e.g., `i2c_address=0x50`)
            i2c_bus (int): The I2C bus to use. Should probably be set to 1 on Raspberry Pi.
            pig (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        super().__init__(pig)
        self._i2c_bus = i2c_bus
        self._open(i2c_bus, i2c_address)

    @gpio_command
    def _open(self, i2c_bus, i2c_address):
        """ Opens i2c connection given i2c bus and address."""
        self._handle = self.gpio.i2c_open(i2c_bus, i2c_address)

    @gpio_command
    def _close(self):
        """ Extends superclass method. Checks that pigpiod is connected
        and if a handle has been set - if so, closes an i2c connection.
        """
        super()._close()
        self.gpio.i2c_close(self.handle)

    @gpio_command
    def read_device(self, count=2) -> tuple:
        """ Read a specified number of bytes directly from the the device without specifying or changing the register.
        Does NOT perform LE/BE conversion.

        Args:
            count (int): The number of bytes to read from the device.

        Returns:
            tuple: a tuple of the number of bytes read and a bytearray containing the bytes. If there was an error the
            number of bytes read will be less than zero (and will contain the error code).
        """
        return self.gpio.i2c_read_device(self.handle, count)

    @gpio_command
    def write_device(self, word, signed=False):
        """ Write 2 bytes to the device without specifying register. DOES perform LE/BE conversion.

        Args:
            word (int): The integer representation of the data to write.
            signed (bool): Whether or not `word` is signed.
        """
        self.gpio.i2c_write_device(
            self.handle,
            native16_to_be(word, signed=signed)
        )

    @gpio_command
    def read_register(self, register, signed=False) -> int:
        """ Read 2 bytes from the specified register and byteswap the result.

        Args:
            register (int): The index of the register to read.
            signed (bool): Whether or not the data to read is expected to be signed.

        Returns:
            int: integer representation of 16 bit register contents.
        """
        return be16_to_native(self.gpio.i2c_read_i2c_block_data(
            self.handle,
            register,
            count=2
        ), signed=signed)

    @gpio_command
    def write_register(self, register, word, signed=False):
        """ Write 2 bytes to the specified register. Byteswaps.

        Args:
            register (int): The index of the register to write to
            word (int): The unsigned 16 bit integer to write to the register (must be consistent with 'signed')
            signed (bool): Whether or not 'word' is signed
        """
        self.gpio.i2c_write_i2c_block_data(
            self.handle,
            register,
            native16_to_be(word, signed=signed)
        )

    class Register:
        """ Describes a writable configuration register. Has dynamically
        defined attributes corresponding to the fields described by the
        passed arguments. Takes as arguments two tuples of equal length,
        the first of which names each field and the second being a tuple
        of tuples containing the (human readable) possible settings &
        values for each field.

        Note: The initializer reverses the fields & their values because
        a human reads the register, as drawn in the datasheet, from left
        to right - however, the fields furthest to the left are the most
        significant bits of the register.
        """

        def __init__(self, fields, values):
            """ Initializer which loads (dynamically defined) attributes from tuples.

            Args:
                fields (tuple): A tuple containing the names of the register's value fields
                values (tuple): A tuple of tuples containing the possible values for each value field. Length must match
                    the length of fields. If there are redundant values for a field specified in the datasheet, be sure
                    to include them. (e.g., a field takes values `A: 0b00`, `B: 0b01`, and `C: 0b10`; but the value for
                    `0b11` is either not specified by the datasheet or is listed redundantly as `C: 0b11` -> `values`
                    should list both the 3rd and 4th possible values as 'C' like so: ('A', 'B', 'C', 'C')
            """
            if len(fields) != len(values):
                raise ValueError('fields and values must contain the same number of elements')
            self.fields = fields
            offset = 0
            for fld, val in zip(reversed(fields), reversed(values)):
                setattr(
                    self,
                    fld,
                    self.ValueField(
                        offset,
                        len(val) - 1,
                        OrderedDict(zip(val, range(len(val))))
                    )
                )
                offset += (len(val) - 1).bit_length()

        def unpack(self, cfg) -> OrderedDict:
            """ Given the contents of a register in integer form, returns a dict of fields and their current settings.

            Args:
                cfg (int): An integer representing a possible configuration value for the register
            """
            return OrderedDict(zip(
                self.fields,
                (getattr(
                    getattr(self, field),
                    'unpack')(cfg) for field in self.fields)
            ))

        def pack(self, cfg, **kwargs) -> int:
            """ Given an initial integer representation of a register and an arbitrary number of field=value settings,
            returns an integer representation of the register incorporating the new settings.

            Args:
                cfg (int): An integer representing a possible configuration value for the register
                **kwargs: The register fields & values to patch into cfg. Takes keyword arguments of the form:
                    `field=value`
            """
            for field, value in kwargs.items():
                if hasattr(self, field) and value is not None:
                    cfg = getattr(getattr(self, field), 'insert')(cfg, value)
            return cfg

        class ValueField:
            """ Describes a configurable value field in a writable register."""

            def __init__(self, offset, mask, values):
                """ Instantiates a value field of a register given the bit offset, mask, and list of possible values.

                Args:
                    offset (int): The offset bits of the value field in the register, i.e. the distance from LSB
                    mask (int): integer representation of the value field mask (w/o offset)
                    values (OrderedDict): The possible values that the field can take.
                """
                self._offset = offset
                self._mask = mask
                self._values = values
                self._reversed_values = OrderedDict(map(reversed, self._values.items()))

            def unpack(self, cfg):
                """ Extracts the ValueField's setting from cfg & returns the result in a human readable form.

                Args:
                    cfg (int): An integer representing a possible configuration value for the register
                """
                return self._reversed_values[self.extract(cfg)]

            def extract(self, cfg) -> int:
                """ Extracts setting from passed 16-bit config & returns integer representation.

                Args:
                    cfg (int): An integer representing a possible configuration value for the register
                """
                return (cfg & (self._mask << self._offset)) >> self._offset

            def pack(self, value) -> int:
                """ Takes a human-readable ValueField setting and returns the corresponding bit-shifted integer.

                Args:
                    value (int): The integer representation of `value` bit-shifted by the ValueField's offset

                Returns:
                    int: The integer representation of the ValueField setting according to `value`
                """
                if value not in self._values.keys():
                    raise ValueError("ValueField must be one of: {}".format(self._values.keys()))
                return self._values[value] << self._offset

            def insert(self, cfg, value) -> int:
                """ Validates and performs bitwise replacement with the human-readable ValueField setting and integer
                representation of the register configuration.

                Args:
                    cfg (int): An integer representing a possible configuration value for the register
                    value (object): The human readable representation of the desired ValueField setting. Must match a
                        value in ValueField._values; if not, throws a ValueError

                Returns:
                    int: The integer representation of the Register's configuration with the value of ValueField patched
                        according the `value`
                """
                if value not in self._values.keys():
                    raise ValueError("ValueField must be one of: {}".format(self._values.keys()))
                return (cfg & ~(self._mask << self._offset)) | (self._values[value] << self._offset)


class ADS1015(I2CDevice):
    """ ADS1015 12 bit, 4 Channel Analog to Digital Converter.
    Datasheet:
        http://www.ti.com/lit/ds/symlink/ads1015.pdf?&ts=1589228228921

    Default Values:
     Default configuration for vent:     0xC3E3
     Default configuration on power-up:  0x8583
    """
    _DEFAULT_ADDRESS = 0x48
    _DEFAULT_VALUES = {'MUX': 0, 'PGA': 4.096, 'MODE': 'SINGLE', 'DR': 3300}
    _TIMEOUT = 1
    """ Address Pointer Register (write-only) """
    _POINTER_FIELDS = ('P',)
    _POINTER_VALUES = (
        (
            'CONVERSION',
            'CONFIG',
            'LO_THRESH',
            'HIGH_THRESH'
        ),
    )

    """ Config Register (R/W) """
    _CONFIG_FIELDS = (
        'OS',
        'MUX',
        'PGA',
        'MODE',
        'DR',
        'COMP_MODE',
        'COMP_POL',
        'COMP_LAT',
        'COMP_QUE'
    )
    _CONFIG_VALUES = (
        ('NO_EFFECT', 'START_CONVERSION'),
        ((0, 1), (0, 3), (1, 3), (2, 3), 0, 1, 2, 3),
        (6.144, 4.096, 2.048, 1.024, 0.512, 0.256, 0.256, 0.256),
        ('CONTINUOUS', 'SINGLE'),
        (128, 250, 490, 920, 1600, 2400, 3300, 3300),
        ('TRADIONAL', 'WINDOW'),
        ('ACTIVE_LOW', 'ACTIVE_HIGH'),
        ('NONLATCHING', 'LATCHING'),
        (1, 2, 3, 'DISABLE')
    )
    USER_CONFIGURABLE_FIELDS = ('MUX', 'PGA', 'MODE', 'DR')
    """ Note:
    The Conversion Register is read-only and contains a 16bit
    representation of the requested value (provided the conversion is
    ready).

    The Lo-thresh & Hi-thresh Registers are not Utilized here. However,
    their function and usage are described in the datasheet. Should you
    want to extend the functionality implemented here.
    """

    def __init__(self, address=_DEFAULT_ADDRESS, i2c_bus=1, gpio=None):
        """ Initializes registers: Pointer register is write only,
        config is R/W. Sets initial value of _last_cfg to what is
        actually on the ADS.Packs default settings into _cfg, but does
        not actually write to ADC - that occurs when read_conversion()
        is called.

        Args:
            address (int): I2C address of the device. (e.g., `i2c_address=0x48`)
            i2c_bus (int): The I2C bus to use. Should probably be set to 1 on Raspberry Pi.
            gpio (PigpioConnection): pigpiod connection to use; if not specified, a new one is established
        """
        super().__init__(address, i2c_bus, gpio)
        self.pointer = self.Register(self._POINTER_FIELDS, self._POINTER_VALUES)
        self._config = self.Register(self._CONFIG_FIELDS, self._CONFIG_VALUES)
        self._last_cfg = self._read_last_cfg()
        self._cfg = self._config.pack(cfg=self._last_cfg, **self._DEFAULT_VALUES)

    def read_conversion(self, **kwargs) -> float:
        """ Returns a voltage (expressed as a float) corresponding to a channel on the ADC.
        The channel to read from, along with the gain, mode, and sample rate of the conversion may be may be  specified
        as optional parameters. If read_conversion() is called with no parameters, the resulting voltage corresponds to
        the channel last read from and the same conversion settings.

        Args:
            MUX: The pin to read from in single channel mode: e.g., `0, 1, 2, 3`
                or, a tuple of pins over which to make a differential reading.
                e.g., `(0, 1), (0, 3), (1, 3), (2, 3)`
            PGA: The full scale voltage (FSV) corresponding to a programmable gain setting.
                e.g., `(6.144, 4.096, 2.048, 1.024, 0.512, 0.256, 0.256, 0.256)`
            MODE: Whether to set the ADC to continuous conversion mode, or operate in single-shot mode.
                e.g., `'CONTINUOUS', 'SINGLE'`
            DR: The data rate to make the conversion at; units: samples per second.
                e.g., `8, 16, 32, 64, 128, 250, 475, 860`
        """
        return (
                self._read_conversion(**kwargs)
                * self.config.PGA.unpack(self.cfg) / 32767
        )

    def print_config(self) -> OrderedDict:
        """ Returns the human-readable configuration for the next read.

        Returns:
            OrderedDict: an ordered dictionary of the form {field: value}, ordered from MSB -> LSB
        """
        return self.config.unpack(self.cfg)

    @property
    def config(self):
        """ Returns the Register object of the config register.

        Returns:
            vent.io.devices.I2CDevice.Register: The Register object initialized for the ADS1115.
        """
        return self._config

    @property
    def cfg(self) -> int:
        """ Returns the contents (as a 16-bit unsigned integer) of the configuration that will be written to the config
        register when read_conversion() is next called.
        """
        return self._cfg

    def _read_conversion(self, **kwargs) -> int:
        """ Backend for read_conversion. Returns the contents of the 16-bit conversion register as an unsigned integer.

        If no parameters are passed, one of two things can happen:

            1)  If the ADC is in single-shot (mode='SINGLE') conversion
                mode, _last_cfg is written to the config register; once
                the ADC indicates it is ready, the contents of the
                conversion register are read and the result is returned.
            2)  If the ADC is in CONTINUOUS mode, the contents of the
                conversion register are read immediately and returned.

        If any of channel, gain, mode, or data_rate are specified as
        parameters, a new _cfg is packed and written to the config
        register; once the ADC indicates it is ready, the contents of
        the conversion register are read and the result is returned.

        Note: In continuous mode, data can be read from the conversion
        register of the ADS1115 at any time and always reflects the
        most recently completed conversion. So says the datasheet.

        Args:
            **kwargs: see documentation of vent.io.devices.ADS1115.read_conversion
        """
        self._cfg = self._config.pack(cfg=self.cfg, **kwargs)
        mode = self.print_config()['MODE']
        if self._cfg != self._last_cfg or mode == 'SINGLE':
            self.write_register(self.pointer.P.pack('CONFIG'), self.cfg)
            self._last_cfg = self.cfg
            data_rate = self._config.DR.unpack(self.cfg)
            '''while not (self._ready() or mode == 'CONTINUOUS'):
                # TODO: Needs timout
                tick = time.time()
                while (time.time() - tick) < (1 / data_rate):
                    pass  # TODO: implement asyncio.sleep()'''
        return self.read_register(self.pointer.P.pack('CONVERSION'), signed=True)

    def _read_last_cfg(self) -> int:
        """ Reads the config register and returns the contents as a 16-bit unsigned integer;
        updates internal record _last_cfg.
        """
        self._last_cfg = self.read_register(self.pointer.P.pack('CONFIG'))
        return self._last_cfg

    def _ready(self) -> bool:
        """ Return status of ADC conversion; True indicates the conversion is complete and the results ready to be read.
        """
        return bool(self.read_register(self.pointer.P.pack('CONFIG')) >> 15)


def be16_to_native(data, signed=False) -> int:
    """ Unpacks a bytes-like object respecting big-endianness of outside world and returns an int according to signed.

    Args:
        data: bytes-like object. The data to be unpacked & converted
        signed (bool): Whether or not `data` is signed
    """
    return int.from_bytes(data[1][:2], 'big', signed=signed)


def native16_to_be(word, signed=False) -> bytes:
    """ Packs an int into bytes after swapping endianness.

    Args:
        signed (bool): Whether or not `data` is signed
        word (int): The integer representation to converted and packed into bytes
    """
    return word.to_bytes(2, 'big', signed=signed)



import os
import numpy as np
import torch

DEFAULT_PRESSURE_RANGE = (5.0, 35.0)
DEFAULT_KEYPOINTS = [1e-8, 1.0, 1.5, 3.0]
DEFAULT_BPM = 20


class BreathWaveform:
    """Waveform generator with shape |‾\_"""

    def __init__(self, range=None, keypoints=None, bpm=DEFAULT_BPM, kernel=None, dt=0.01):
        self.range = range
        self.lo, self.hi = range or DEFAULT_PRESSURE_RANGE
        self.fp = [self.lo, self.hi, self.hi, self.lo, self.lo]

        self.xp = np.zeros(len(self.fp))
        self.xp[1:] = np.array(keypoints or DEFAULT_KEYPOINTS)
        self.xp[-1] = 60 / bpm

        self._keypoints = self.xp

        pad = 0
        num = int(1 / dt)
        if kernel is not None:
            pad = 60 / bpm / (num - 1)
            num += len(kernel) // 2 * 2

        tt = np.linspace(-pad, 60 / bpm + pad, num)
        self.fp = self.at(tt)
        self.xp = tt

        if kernel is not None:
            self.fp = np.convolve(self.fp, kernel, mode="valid")
            self.xp = np.linspace(0, 60 / bpm, int(1 / dt))

    @property
    def keypoints(self):
        if hasattr(self, "_keypoints"):
            return self._keypoints
        else:
            return self.xp

    @property
    def period(self):
        return self.xp[-1]

    @property
    def PIP(self):
        if hasattr(self, "range"):
            return self.range[1]
        else:
            return np.max(self.fp)

    @property
    def PEEP(self):
        if hasattr(self, "range"):
            return self.range[0]
        else:
            return np.min(self.fp)

    def at(self, t):
        return np.interp(t, self.xp, self.fp, period=self.period)

    def elapsed(self, t):
        return t % self.period

    def decay(self, t):
        elapsed = self.elapsed(t)
        if elapsed < self.keypoints[2]:
            return None
        elif elapsed < self.keypoints[3]:
            return 0.0
        else:
            return 5 * (1 - np.exp(5 * (self.keypoints[3] - elapsed)))

    def phase(self, t):
        return np.searchsorted(self.keypoints, t % self.period, side="right")


class TorchStandardScaler:
    def __init__(self, scaler=None, dtype=None, device="cpu"):
        if dtype is None:
            torch.get_default_dtype()
        if scaler is not None:
            self.mean = torch.tensor(scaler.mean_, dtype=dtype, device=device)
            self.std = torch.tensor(scaler.scale_, dtype=dtype, device=device)

    def fit(self, x):
        self.mean = x.mean(0, keepdim=True)
        self.std = x.std(0, unbiased=False, keepdim=True)

    def transform(self, x):
        return (x - self.mean) / self.std

    def inverse_transform(self, x):
        return x * self.std + self.mean

    def to(self, device):
        self.mean.to(device)
        self.std.to(device)



import numpy as np
import torch
from copy import deepcopy

from deluca.lung.environments.core import Environment
from deluca.lung.utils.core import TorchStandardScaler


class StitchedSim(Environment):
    def __init__(self, inspiratory_model, u_window, p_window, u_scaler, p_scaler, to_round=False):
        super().__init__()

        self.inspiratory_model = inspiratory_model

        self.u_window = u_window
        self.p_window = p_window

        self.u_scaler = TorchStandardScaler(u_scaler)
        self.p_scaler = TorchStandardScaler(p_scaler)

        self.to_round = to_round

        self.reset()

    def reset(self):
        self.scaled_peep = self.tensor(self.inspiratory_model.boundary_dict["0"]().item())
        self.state = {"t_in": 0, "u_history": [], "p_history": [self.scaled_peep]}
        self.sync_unscaled_pressure()

    def set_current_pressure(self, pressure):
        self.state["p_history"][-1] = self.p_scaler.transform(pressure.detach().clone()).squeeze()
        self.sync_unscaled_pressure()

    def sync_unscaled_pressure(self):
        self.pressure = self.p_scaler.inverse_transform(self.state["p_history"][-1]).squeeze()

    def copy_state(self, state):
        new_state = {}
        new_state["t_in"] = state["t_in"]
        new_state["u_history"] = [u.detach().clone() for u in state["u_history"]]
        new_state["p_history"] = [u.detach().clone() for u in state["p_history"]]
        return new_state

    def cache_state(self, return_state=False):
        # set up a snapshot for rewind_state
        state_copy = self.copy_state(self.state)
        if return_state:
            return state_copy
        else:
            self.cached_state = state_copy

    def rewind_state(self, state=None):
        # load a history state cached by cache_state
        if state:
            self.state = self.copy_state(state)
        else:
            self.state = self.copy_state(self.cached_state)
        self.sync_unscaled_pressure()

    def forward(self, u_in, u_out, t):

        if u_out == 1:
            if self.state["t_in"] > 0:  # reset once per u_out=1
                self.state = {"t_in": 0, "u_history": [], "p_history": [self.scaled_peep]}
                self.sync_unscaled_pressure()

        else:
            if self.to_round:
                u_in_scaled = self.u_scaler.transform(torch.round(u_in)).squeeze()
            else:
                u_in_scaled = self.u_scaler.transform(u_in).squeeze()

            self.state["u_history"].append(u_in_scaled)

            self.state["t_in"] += 1
            t_key = str(self.state["t_in"])

            if t_key in self.inspiratory_model.boundary_dict:  # predict from boundary model
                features = torch.cat(
                    [
                        torch.stack(self.state["u_history"]),
                        torch.stack(self.state["p_history"]),
                    ]
                )
                scaled_pressure = self.inspiratory_model.boundary_dict[t_key](features)
            else:  # predict from default model
                features = torch.cat(
                    [
                        torch.stack(self.state["u_history"][-self.u_window :]),
                        torch.stack(self.state["p_history"][-self.p_window :]),
                    ]
                )
                scaled_pressure = self.inspiratory_model.default_model(features)

            self.state["p_history"].append(scaled_pressure.squeeze())
            self.sync_unscaled_pressure()



from deluca.lung.controllers.core import Controller


class Predestined(Controller):
    def __init__(self, u_ins, u_outs):
        self.u_ins = u_ins
        self.u_outs = u_outs

        self.reset()

    def reset(self):
        self.t = 0

    def compute_action(self, state, t):
        u_in, u_out = self.u_ins[self.t], self.u_outs[self.t]
        self.t += 1

        return (u_in, u_out)



from deluca.lung.controllers.core import Controller

class Impulse(Controller):
    def __init__(self, impulse=50, start=0.5, end=0.65):
        self.impulse = impulse
        self.start = start
        self.end = end
        
    def compute_action(self, state, t):
        u_out = self.u_out(t)

        impulse = 0
        if t >= self.start and t <= self.end:
            impulse = self.impulse

        impulse = impulse if u_out == 0 else 0
            
        return impulse, 0



import os
import torch

from deluca.lung.hal import Hal
from deluca.lung.controllers.core import Controller
from deluca.lung.controllers.core import ControllerRegistry
from deluca.lung.environments.core import Environment
from deluca.lung.environments.core import EnvironmentRegistry
from deluca.lung.utils import BreathWaveform
from deluca.lung.utils.experiment import experiment

ROOT = os.path.dirname(os.path.realpath(os.path.join(__file__, "..")))

# NOTE: We set this here for convenience, but should think of a better way to
# do this
torch.manual_seed(0)

torch.set_default_dtype(torch.float64)

__all__ = [
    "Hal",
    "Controller",
    "ControllerRegistry",
    "Environment",
    "EnvironmentRegistry",
    "BreathWaveform",
    "experiment",
]



from deluca.lungimental.explorer import explore
from deluca.lungimental.pid_grid import pid_grid
from deluca.lungimental.best_pid import best_pid
from deluca.lungimental.run_controller import run_controller
from deluca.lungimental.core import run_calibration, save_data_and_plot



import os
import numpy as np
import matplotlib
import dill as pickle
import matplotlib.pyplot as plt

from deluca.lung.utils import BreathWaveform


class Analyzer:
    def __init__(self, path):
        self.data = path if isinstance(path, dict) else pickle.load(open(path, "rb"))
        self.controller = self.data["controller"]
        timeseries = self.data["timeseries"]
        self.tt = timeseries["timestamp"]
        self.u_in = timeseries["u_in"]
        self.u_out = timeseries["u_out"]
        self.pressure = timeseries["pressure"]
        self.target = timeseries["target"]
        self.flow = timeseries["flow"]

    ###########################################################################
    # Plotting methods
    ###########################################################################

    def plot(
        self,
        title=None,
        axes=None,
        figsize=None,
        xlim=None,
        ylim=[0, 60],
        legend=False,
        control=True,
        **kwargs
    ):
        matplotlib.pyplot.figure(figsize=(figsize or (12, 6)))
        # trash
        if axes is None:
            axes = matplotlib.pyplot.axes()

        if xlim is not None:
            axes.set_xlim(xlim)

        axes.set_xlabel("Time (s)")
        plts = []
        (pressure,) = axes.plot(
            self.tt, self.pressure, color="blue", label="actual pressure", **kwargs
        )
        (target,) = axes.plot(
            self.tt, self.target, color="orange", label="target pressure", **kwargs
        )
        axes.set_ylabel("Pressure (cmH2O)")
        axes.set_ylim(ylim)
        expiratory = axes.fill_between(
            self.tt,
            axes.get_ylim()[0],
            axes.get_ylim()[1],
            where=self.u_out.astype(bool),
            color="dimgray",
            alpha=0.3,
            label="expiratory",
        )
        inspiratory = axes.fill_between(
            self.tt,
            axes.get_ylim()[0],
            axes.get_ylim()[1],
            where=np.logical_not(self.u_out.astype(bool)),
            color="lightgray",
            alpha=0.3,
            label="inspiratory",
        )
        plts.extend([pressure, target, inspiratory, expiratory])

        if control:
            twin_ax = axes.twinx()
            twin_ax.set_ylim([-2, 102])
            (u_in,) = twin_ax.plot(
                self.tt, np.clip(self.u_in, 0, 100), c="gray", label="u_in", **kwargs
            )
            twin_ax.set_ylabel("Control")
            plts.append(u_in)

        if title is not None:
            plt.title(title)

        if legend:
            labels = [p.get_label() for p in plts]
            plt.legend(
                plts,
                labels,
                bbox_to_anchor=(-.05, 1.02, 1.1, 0.05),
                mode="expand",
                ncol=4,
                loc="lower left",
            )

    def plot_inspiratory_clips(self, **kwargs):
        inspiratory_clips = self.infer_inspiratory_phases()

        plt.subplot(121)
        plt.title("u_in")
        for start, end in inspiratory_clips:
            u_in = self.u_in[start:end]
            plt.plot(self.tt[start:end] - self.tt[start], u_in, "k", alpha=0.1)

        plt.subplot(122)
        plt.title("pressure")
        for start, end in inspiratory_clips:
            pressure = self.pressure[start:end]
            plt.plot(self.tt[start:end] - self.tt[start], pressure, "b", alpha=0.1)

    ###########################################################################
    # Utility methods
    ###########################################################################

    def infer_inspiratory_phases(self, use_cached=True):
        # finds inspiratory phase intervals from expiratory valve controls
        # returns list of endpoints so that u_out[lo:hi] == 1

        if not use_cached or not hasattr(self, "cached_inspiratory_phases"):
            d_u_out = np.diff(self.u_out, prepend=1)

            starts = np.where(d_u_out == -1)[0]
            ends = np.where(d_u_out == 1)[0]

            self.cached_inspiratory_phases = list(zip(starts, ends))

        return self.cached_inspiratory_phases

    ###########################################################################
    # Metric methods
    ###########################################################################

    def losses_per_breath(self, target, loss_fn=None):
        # computes trapezoidally integrated loss per inferred breath

        loss_fn = loss_fn or np.square

        # handle polymorphic targets
        if isinstance(target, int):
            target_fn = lambda _: target
        elif isinstance(target, BreathWaveform):
            target_fn = lambda t: target.at(t)
        else:
            raise ValueError("unrecognized type for target")

        breaths = self.infer_inspiratory_phases()
        losses = []

        # integrate loss for each detected inspiratory phase
        for start, end in breaths:
            errs = loss_fn(target_fn(self.tt[start:end]) - self.pressure[start:end])
            loss = np.trapz(errs, self.tt[start:end])
            losses.append(loss)

        return np.array(losses)

    def default_metric(self, target=None, loss_fn=np.abs):
        # I suggest keeping a separate function for default settings
        # so nobody has to change code if we change the benchmark
        # as it stands: average loss across breaths, discounting first breath
        target = target or self.controller.waveform

        return self.losses_per_breath(target, loss_fn)[1:].mean()



from setuptools import setup, find_namespace_packages

setup(
    name="deluca-lung",
    version="0.0.1",
    url="https://github.com/MinRegret/deluca-lung",
    author="Google AI Princeton",
    author_email="dsuo@google.com",
    description="Ventilator control research in deluca",
    packages=find_namespace_packages(),
    install_requires=[
        "numpy",
        "tqdm",
        "torch",
        "scipy",
        "sklearn",
        "matplotlib",
        "jupyter",
        "ipython",
        "pathos",
        "pigpio",
        "click",
    ],
)


from deluca.lungimental.controllers._ibonus import IBonus, IBonusCubic, LookaheadPID
from deluca.lungimental.controllers._deep_pid_residual import DeepPIDResidual
from deluca.lungimental.controllers._deep_pid_residual_clipped import DeepPIDResidualClipped
from deluca.lungimental.controllers._pid_correction import PIDCorrection


import os
import pickle

import torch
import numpy as np
import matplotlib.pyplot as plt

from deluca.lungollers import ResidualExplorer, PID
from deluca.lungonments import PhysicalLung
from deluca.lung.analyzer import Analyzer
from deluca.lung import BreathWaveform
from deluca.lungimental.core import run_calibration


plt.rc("figure", figsize=(10, 3))


def collect_runs(controller, path, n_runs=1, append_to=None, vent=None, **kwargs):
    device = None if isinstance(vent, PhysicalLung) else "cpu"
    if append_to is None:
        results = []
    else:
        results = append_to

    for i in range(n_runs):
        vent.reset()
        result = vent.run(controller, device=device, **kwargs)
        results.append(result)
        pickle.dump(result, open(path % i, "wb"))
    return results


def explore(
    controller,
    directory,
    R=50,
    C=10,
    PEEP=5,
    T=10000,
    n_runs=5,
    PIPs=[10, 15, 20, 25, 30, 35],
    abort=70,
    dt=0.03,
    vent=None,
):
    if not isinstance(controller, ResidualExplorer):
        print("ERROR: expecting ResidualExplorer as controller")
        return

    vent = vent or PhysicalLung()
    print("Running calibration")
    run_calibration(R, C, PEEP, directory)

    all_results = {}
    for PIP in PIPs:
        path = os.path.join(directory, f"R{R}C{C}PEEP{PEEP}_PIP{PIP}_pid_triangle_residual_%i.pkl")

        controller.update_waveform(BreathWaveform((PEEP, PIP)))

        results = collect_runs(
            controller,
            path,
            R=R,
            C=C,
            PEEP=PEEP,
            T=T,
            dt=dt,
            n_runs=n_runs,
            abort=abort,
            vent=vent,
        )
        all_results[PIP] = results

    print("Running calibration")
    run_calibration(R, C, PEEP, directory)

    return all_results



from deluca.lung.utils.core import BreathWaveform
from deluca.lung.utils.core import TorchStandardScaler
from deluca.lung.utils.analyzer import Analyzer
from deluca.lung.utils.munger import Munger
from deluca.lung.utils.nn import SNN
from deluca.lung.utils.nn import BoundaryDict
from deluca.lung.utils.nn import ShallowBoundaryModel
from deluca.lung.utils.nn import RegressionBoundaryModel
from deluca.lung.utils.nn import ConstantModel
from deluca.lung.utils.nn import InspiratoryModel

__all__ = [
    "BreathWaveform",
    "TorchStandardScaler",
    "Munger",
    "Analyzer",
    "Munger",
    "SNN",
    "BoundaryDict",
    "ShallowBoundaryModel",
    "RegressionBoundaryModel",
    "ConstantModel",
    "InspiratoryModel"
]



from deluca.lung.utils.analyzer import Analyzer
from deluca.lung.utils.munger import Munger
from deluca.lung.controllers._predestined import Predestined

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import torch

def open_loop_test(sim, munger_path, key="test", abort=70, use_tqdm=False, **kwargs):
    # run u_ins, get pressures, check with true trajectory

    munger = Munger.load(munger_path)

    test_summary = {}

    all_runs = []

    for test_trajectory in munger.splits[key]:
        test_u_ins, test_pressures = test_trajectory
        T = len(test_u_ins)

        controller = Predestined(test_u_ins, np.zeros_like(test_u_ins))

        run_data = sim.run(
            controller, T=T, abort=abort, use_tqdm=use_tqdm, **kwargs
        )

        analyzer = Analyzer(run_data)

        # pedantic copying + self-documenting names...
        preds = analyzer.pressure.copy()
        truth = test_pressures.copy()

        all_runs.append((preds, truth))

    test_summary['all_runs'] = all_runs
    populate_resids(test_summary)

    return test_summary

def forecasting_test(
    sim, munger_path, key="test", abort=70, use_tqdm=False, **kwargs
):
    munger = Munger.load(munger_path)

    test_summary = {}
    all_runs = []

    with torch.no_grad():
        for test_trajectory in munger.splits[key]:
            test_u_ins, test_pressures = test_trajectory
            test_u_ins = sim.tensor(test_u_ins)
            test_pressures_tensor = sim.tensor(test_pressures)

            T = len(test_u_ins)

            for k in range(T):  # feed k steps of pressure history
                sim.reset()
                sim.state["t_in"] = max(k-1, 0)
                if k > 0:
                    sim.state["u_history"] = list(sim.u_scaler.transform(test_u_ins[:k-1]))
                    sim.state["p_history"] = list(sim.p_scaler.transform(test_pressures_tensor[:k]))
                    sim.sync_unscaled_pressure()

                forecasts = []  # predicted pressure trajectory

                for t in range(k-1, T):  # feed the rest from generation
                    if t == -1:  # with no pressure feedback, p[0] is a prediction
                        forecasts.append(sim.pressure.item())
                    else:  # predict next p
                        u_in = test_u_ins[t]
                        sim(u_in, 0, t)
                        if t < T-1:  # last forecast is available but not benchmarkable
                            forecasts.append(sim.pressure.item())

                # model's k-step forecast residuals
                preds = np.array(forecasts)
                truth = test_pressures[k:].copy()
                all_runs.append((preds, truth))

    test_summary['all_runs'] = all_runs
    populate_resids(test_summary)

    return test_summary

def populate_resids(test_summary, key='all_runs', l1_out='mae', l2_out='rmse'):
    all_runs = test_summary[key]

    Tmax = max(map(lambda x:len(x[0]), all_runs))
    l1, l2, counts = np.zeros(Tmax), np.zeros(Tmax), np.zeros(Tmax)

    for preds, truth in all_runs:
        resids = truth - preds
        T = len(resids)
        l1[:T] += abs(resids)
        l2[:T] += resids**2
        counts[:T] += 1

    test_summary[l1_out] = l1 / counts
    test_summary[l2_out] = (l2 / counts)**0.5

def collate_boundary(all_runs, first_k=5):
    all_preds = []
    all_truth = []
    
    for preds, truth in all_runs:
        if len(preds) < first_k:
            continue
        all_preds.append(preds[:first_k])
        all_truth.append(truth[:first_k])

    return np.array(all_preds), np.array(all_truth)

def plot_spray(all_runs, **kwargs):
    if len(all_runs) < 1000:
        indices = np.arange(len(all_runs))
    else:  # subsample if there are too many runs
        indices = np.random.choice(len(all_runs), 1000)

    for i in indices:
        preds, truth = all_runs[i]
        plt.plot(truth - preds, 'gray', alpha=0.1)

    plt.xlabel('# simulated pressures')
    plt.grid()

def plot_error_metric(errors, **kwargs):
    plt.plot(errors, 'b', marker='o', markersize=4, lw=1, **kwargs)
    for i,err in enumerate(errors):
        if i%2 == 0:
            plt.annotate(f"{err:.2f}", (i,err+0.2), color='b', ha='center', va='bottom')
        else:
            plt.annotate(f"{err:.2f}", (i,err-0.2), color='b', ha='center', va='top')

def plot_sim_diagnostics(open_loop_summary=None, forecast_summary=None, run=None):
    plt.rc('figure', figsize=(8, 8))

    plt.subplot(211)  ##### open-loop test
    for preds, truth in open_loop_summary['all_runs']:
        plt.plot(truth - preds, 'gray', alpha=0.1)
    plt.title('open-loop test')
    plt.ylabel('residual')
    plot_spray(open_loop_summary['all_runs'])
    plot_error_metric(open_loop_summary['mae'])

#     plt.subplot(132)  ##### forecasting test
#     plt.title('forecasting test')
#     plot_spray(forecast_summary['all_runs'])
#     plot_error_metric(forecast_summary['mae'])
    
    plt.subplot(212)
    plt.title('pressure trajectories')
    plot_run(run)
    

    #### boundary residual plots
#     boundary_preds, boundary_truth = collate_boundary(open_loop_summary['all_runs'])
#     for i in range(5):
#         plt.subplot(2,5,6+i)
#         if i == 0:
#             plt.ylabel('predicted pressure')
#         plt.xlabel(f'true pressure (t={i+1})')
#         plt.scatter(boundary_truth[:,i], boundary_preds[:,i], s=2, color='b')
#         plt.grid()

#         x_min, x_max = plt.xlim()
#         y_min, y_max = plt.ylim()
#         z = [min(x_min,y_min), max(x_max,y_max)]
#         plt.plot(z, z, 'gray')

    plt.tight_layout()

def default_sim_diagnostics(sim, munger_path, open_loop_summary=None, forecast_summary=None):
    # full eval protocol, for the exceptionally lazy
    open_loop_summary = open_loop_summary or open_loop_test(sim, munger_path)
    forecast_summary = forecast_summary or forecasting_test(sim, munger_path)

    plot_sim_diagnostics(open_loop_summary, forecast_summary)

def plot_intrinsic_uncertainty(munger_path, figsize=(4, 4), x1_idx=("u", 0), x2_idx=None, y_idx=("p", 1), key="test", s=5, alpha=0.85):
    munger = Munger.load(munger_path)
    plt.rc('figure', figsize=figsize)

    x1 = []
    x2 = []
    y = []

    for trajectory in munger.splits[key]:
        u, p = trajectory
        x1.append(locals()[x1_idx[0]][x1_idx[1]])
        if x2_idx is not None:
            x2.append(locals()[x2_idx[0]][x2_idx[1]])
        y.append(locals()[y_idx[0]][y_idx[1]])

    if x2_idx is None:
        plt.scatter(x1, y, s=s, alpha=alpha)

        plt.ylabel(f"{y_idx}")
    else:
        # ys = np.zeros((len(x1), len(x2)))
        # for i in x1:
        contours = plt.tricontour(x1, x2, y, s=s, alpha=alpha)
        plt.ylabel(f"{x2_idx}")
        colorbar =  plt.colorbar()
        plt.clabel(contours, inline=True, fontsize=8)
        colorbar.set_label(f"{y_idx}", rotation=90)


    plt.title("Intrinsic uncertainty")
    plt.xlabel(f"{x1_idx}")
    
    
def get_run(sim, munger_path, key="test", abort=70, use_tqdm=False, **kwargs):
    # run u_ins, get pressures, check with true trajectory

    munger = Munger.load(munger_path)
    
    test_trajectory = munger.splits[key][0]

    test_u_ins, test_pressures = test_trajectory
    T = len(test_u_ins)

    controller = Predestined(test_u_ins, np.zeros_like(test_u_ins))

    run_data = sim.run(
        controller, T=T, abort=abort, use_tqdm=use_tqdm, **kwargs
    )

    analyzer = Analyzer(run_data)

    # pedantic copying + self-documenting names...
    preds = analyzer.pressure.copy()
    truth = test_pressures.copy()

    return (preds, truth)

def plot_run(run, **kwargs):
    preds, truth = run
    plt.plot(truth, 'blue', alpha=1.0, label='true trajectory')
    plt.plot(preds, 'orange', alpha=1.0, label='simulated trajectory')
    plt.legend(loc='lower right')

    plt.xlabel('# time steps')
    plt.ylabel('pressure (cm H2O)')
    plt.grid()


from deluca.lungollers.core import Controller
from deluca.lung import BreathWaveform

import numpy as np
import torch

class ClippedDeep(Controller):
    def __init__(
        self,
        clip=35.,
        H=100,
        waveform=None,
        bptt=1,
        input_dim=1,
        activation=torch.nn.ReLU,
        history_len=10,
        kernel_size=5,
        **kwargs
    ):
        super().__init__()

        self.clip = clip
        self.internal_time = 0
        self.last_t = 0

        self.model = torch.nn.Sequential(
            torch.nn.Conv1d(input_dim, H, kernel_size),
            activation(),
            torch.nn.Flatten(),
            torch.nn.Linear(H * (history_len - kernel_size + 1), 1),
        )

        # linear feature transform:
        # errs -> [average of last h errs, ..., average of last 2 errs, last err]
        # emulates low-pass filter bank

        self.featurizer = torch.ones((history_len, history_len), requires_grad=False).tril_()
        self.featurizer /= torch.arange(history_len, 0, -1).unsqueeze(0)

        self.history_len = history_len
        self.input_dim = input_dim
        self.waveform = waveform or BreathWaveform()

        self.breath_start = True

        self.reset()

    def reset(self):
        self.errs = [self.tensor(0.0)] * self.history_len

    def update(self, key, state):
        getattr(self, key).append(state)

    def compute_action(self, state, t):
        self.internal_time += t - self.last_t
        self.last_t = t

        decay = self.waveform.decay(t)

        if decay is None:

            if self.breath_start:
                self.errs = [self.tensor(0.0)] * self.history_len
                self.breath_start = False

            target = self.tensor(self.waveform.at(t))
            self.errs.append(target - state)
            self.errs.pop(0)

            trajectory = torch.stack(self.errs).unsqueeze(0).unsqueeze(0)
            u_in = self.model(trajectory @ self.featurizer)
        else:
            u_in = self.tensor(decay)
            self.breath_start = True
            self.internal_time = 0
            self.last_t = 0

        timed_clip = max(80-100*self.internal_time, self.clip)

        u_in = torch.clamp(u_in, min=0.0, max=timed_clip).squeeze()

        return (u_in, self.u_out(t))

    def train(
        self,
        sim,
        pip_feed="parallel",
        duration=3,
        dt=0.03,
        epochs=100,
        use_noise = False,
        optimizer=torch.optim.Adam,
        optimizer_params={"lr": 1e-3, "weight_decay": 1e-4},
        loss_fn=torch.nn.L1Loss,
        loss_fn_params={},
        scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau,
        scheduler_params={"factor":0.9, "patience":15},
        use_tqdm=True,
        print_loss=1,
        shuffle=False,
        device="cpu",
    ):
        optimizer = optimizer(self.parameters(), **optimizer_params)
        scheduler = scheduler(optimizer, **scheduler_params)
        loss_fn = loss_fn(**loss_fn_params)

        tt = torch.linspace(0, duration, int(duration / dt))
        losses = []

        torch.autograd.set_detect_anomaly(True)

        PIPs = [10, 15, 20, 25, 30, 35]
        PEEP = 5

        # TODO: handle device-awareness
        for epoch in range(epochs):

            if pip_feed == "parallel":
                self.zero_grad()

            for PIP in PIPs:

                if pip_feed == "sequential":
                    self.zero_grad()

                self.waveform = BreathWaveform((PEEP, PIP))

                self.reset()
                sim.reset()

                loss = torch.tensor(0.0, device=device, requires_grad=True)

                for t in tt:
                    sim.pressure += use_noise * torch.normal(mean=torch.tensor(2.), std=1.)
                    pressure = sim.pressure
                    u_in, u_out = self(pressure, t)
                    sim(u_in, u_out, t)

                    if u_out == 0:
                        loss = loss + loss_fn(torch.tensor(self.waveform.at(t)), pressure)

                if pip_feed == "sequential":
                    loss.backward(retain_graph=True)
                    optimizer.step()
                    scheduler.step(loss)
                    per_step_loss = loss / len(tt)
                    losses.append(per_step_loss)

                    if epoch % print_loss == 0:
                        print(
                            f"Epoch: {epoch}, PIP: {PIP}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                        )

            if pip_feed == "parallel":
                loss.backward(retain_graph=True)
                optimizer.step()
                scheduler.step(loss)
                per_step_loss = loss / len(tt)
                losses.append(per_step_loss)

                if epoch % print_loss == 0:
                    print(
                        f"Epoch: {epoch}\tLoss: {per_step_loss:.2f}\tLR: {optimizer.param_groups[0]['lr']}"
                    )

        return losses
